'''
Data and content created by government employees within the scope of their employment are not subject
to domestic copyright protection. 17 U.S.C. 105.

Plugin class method which isolates IXDS html components from primary IXDS to secondary IXDSes

Current use:  Exhibit 26 and EX-FILING FEES attachment documents are processed in secondary IXDSes

Returns nested list of secondary IXDSes and nested lists of modelXbrl.ixdsHtmlElements to be processed by each.
'''

def isolateSeparateIXDSes(modelXbrl, args, **kwargs):
    separateIXDSes = []

    for htmlElt in modelXbrl.ixdsHtmlElements:
        for elt in htmlElt.iterfind(f".//{{{htmlElt.modelDocument.ixNS}}}nonNumeric[@name='dei:DocumentType' or @name='ffd:SubmissnTp']"):
            print(f"trace {elt.tag} = {elt.text}")

    return separateIXDSes



# -*- coding: utf-8 -*-
'''
Custom transforms for SEC

Created by staff of the U.S. Securities and Exchange Commission.
Data and content created by government employees within the scope of their employment
are not subject to domestic copyright protection. 17 U.S.C. 105.

See COPYRIGHT.md for copyright information.

Local copy of text2num.py was obtained from https://github.com/ghewgill/text2num

'''
import datetime
from arelle.ModelValue import qname, lastDayOfMonth
from arelle.Version import authorLabel, copyrightLabel
from arelle.formula.XPathContext import FunctionArgType

#local copy of text2num.py from https://github.com/ghewgill/text2num
from .text2num import text2num, NumberException

from regex import compile as re_compile

# these five transformations take as input a number and output either an exception, or a value in xs:duration type
# they handle zero values and also negative values

# if arg is not an integer, the rest can spill into months and days, but nothing lower
def duryear(arg):
    n, sign = getValue(arg)
    years = int(n)
    months =  (n - years) * 12
    days = int((months - int(months)) * 30.4375)
    months = int(months)
    return durationValue(years, months, days, None, sign)

# if arg is not an integer, the rest can spill into days, but nothing lower
def durmonth(arg):
    n, sign = getValue(arg)
    months = int(n)
    days = int((n - months) * 30.4375)
    return durationValue(None, months, days, None, sign)

# the output will only be in days, nothing lower
# xs:durationType doesn't have weeks, only years, months and days, so we display it all in days
def durweek(arg):
    n, sign = getValue(arg)
    days = int(n * 7)
    return durationValue(None, None, days, None, sign)

# if arg is not an integer, the rest can spill into hours, but nothing lower
def durday(arg):
    n, sign = getValue(arg)
    days = int(n)
    hours = int((n - days) * 24)
    return durationValue(None, None, days, hours, sign)

# the output will only be in hours, nothing lower
def durhour(arg):
    n, sign = getValue(arg)
    hours = int(n)
    return durationValue(None, None, None, hours, sign)

dateQuarterEndPattern = re_compile(r"(^[ \t\n\r]*(?:(?![0-9]{2})\w[^0-9]*)?(1|[Ff]irst|2|[Ss]econd|3|[Tt]hird|4|[Ff]ourth|[Ll]ast)[^0-9]*([0-9]{4})[ \t\n\r]*$)|(^[ \t\n\r]*([0-9]{4})[^0-9]*(1|[Ff]irst|2|[Ss]econd|3|[Tt]hird|4|[Ff]ourth|[Ll]ast)((?![0-9]).*\w+)?[ \t\n\r]*$)")

# the output is a date value of a calendar quarter end
def datequarterend(arg):
    try:
        m = dateQuarterEndPattern.match(arg)
        year = int(m.group(3) or m.group(5)) # year pattern, raises AttributeError if no pattern match
        month = {"first":3, "1":3, "second":6, "2":6, "third":9, "3":9, "fourth":12, "last":12, "4":12
                 }[(m.group(2) or m.group(6)).lower()]
        return str(datetime.date(year, month, lastDayOfMonth(year, month)))
    except (AttributeError, ValueError, KeyError):
        raise FunctionArgType(0, "xs:date")

decimalValuePattern = re_compile(r"^[ \t\n\r]*[+-]?([0-9]+(\.[0-9]*)?|\.[0-9]+)[ \t\n\r]*$")

def getValue(arg):
    if not decimalValuePattern.match(arg):
        raise FunctionArgType(0, "xs:duration")
    try:
        n = float(arg) # could cause a ValueError exception
        if n < 0:
            return (abs(n), '-') # add a negative sign
        return (n, '') # don't add a sign
    except ValueError:
        raise FunctionArgType(0, "xs:duration")

def durationValue(y, m, d, h, sign):

    # preprocess each value so we don't print P0Y0M0D or something like that.
    # in this case, we should print P0Y, and leave out the months and days.
    if all(i == 0 or i is None for i in [y, m, d, h]):
        sign = '' # don't need to print -P0Y, just print P0Y
        hitFirstZeroYet = False
        if y is not None and y == 0:
            hitFirstZeroYet = True
        if m is not None and m == 0:
            if hitFirstZeroYet:
                m = None
            else:
                hitFirstZeroYet = True
        if d is not None and d == 0:
            if hitFirstZeroYet:
                d = None
            else:
                hitFirstZeroYet = True
        if h is not None and h == 0 and hitFirstZeroYet:
            h = None

    output = sign + "P"
    if y is not None:
        output += str(y) + 'Y'
    if m is not None and (m != 0 or y is None):
        output += str(m) + 'M'
    if d is not None and (d != 0 or (y is None and m is None)):
        output += str(d) + 'D'
    if h is not None and (h != 0 or (y is None and m is None and d is None)):
        output += 'T' + str(h) + 'H'
    return output

def numinf(arg):
    return "INF"

def numneginf(arg):
    return "-INF"

def numnan(arg):
    return "NaN"

numwordsenPattern = re_compile(r"^[ \t\n\r]*(((((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)\s+[Hh]undred(\s+(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))\s+[Qq]uintillion(\s*,\s*|\s+|$))?(((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)\s+[Hh]undred(\s+(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))\s+[Qq]uadrillion(\s*,\s*|\s+|$))?(((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)\s+[Hh]undred(\s+(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))\s+[Tt]rillion(\s*,\s*|\s+|$))?(((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)\s+[Hh]undred(\s+(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))\s+[Bb]illion(\s*,\s*|\s+|$))?(((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)\s+[Hh]undred(\s+(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))\s+[Mm]illion(\s*,\s*|\s+|$))?((((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)\s+[Hh]undred(\s+(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))\s+[Tt]housand((\s*,\s*|\s+)((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)\s+[Hh]undred(\s+(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?)))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)\s+[Hh]undred(\s+(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|[Zz]ero|[Nn]o(ne)?|[Nn]il)[ \t\n\r]*$")

numwordsNoPattern = re_compile(r"^[ \t\n\r]*([Nn]o(ne)?|[Nn]il)[ \t\n\r]*$")

commaAndPattern = re_compile(r",|\sand\s") # substitute whitespace for comma or and

def numwordsen(arg):
    if not numwordsenPattern.match(arg) or len(arg) == 0:
        raise FunctionArgType(0, "numwordsen lexical error")
    elif numwordsNoPattern.match(arg): # match "no" or "none"
        return "0"
    try:
        return str(text2num(commaAndPattern.sub(" ", arg.strip().lower()))) # must be returned as a string
    except (NumberException, TypeError, ValueError) as ex:
        raise FunctionArgType(0, str(ex))

durwordsenPattern = re_compile(r"^[ \t\n\r]*((((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)[Hh]undred(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))|[Zz]ero|[Nn]o|[0-9][0-9]{0,3})([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)[Yy]ears?(,?([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)(and\s+)?|$))?((((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)[Hh]undred(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))|[Zz]ero|[Nn]o|[0-9][0-9]{0,3})([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)[Mm]onths?(,?([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)(and\s+)?|$))?((((([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)[Hh]undred(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)(and\s+)?(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))?)|(([Oo]ne|[Tt](wo|hree|en|welve|hirteen)|[Ff](our(teen)?|ive|ifteen)|[Ss](ix(teen)?|even(teen)?)|[Ee](ight(een)?|leven)|[Nn]ine(teen)?)|([Tt](wenty|hirty)|[Ff](orty|ifty)|[Ss](ixty|eventy)|[Ee]ighty|[Nn]inety)(([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)([Oo]ne|[Tt](wo|hree)|[Ff](our|ive)|[Ss](ix|even)|[Ee]ight|[Nn]ine))?))|[Zz]ero|[Nn]o|[0-9][0-9]{0,3})([\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]|\s+)[Dd]ays?)?[ \t\n\r]*$")

durwordZeroNoPattern = re_compile(r"^[ \t\n\r]*[\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]?([Zz]ero|[Nn]o(ne)?)[\u058A\u05BE\u2010\u2011\u2012\u2013\u2014\u2015\uFE58\uFE63\uFF0D-]?[ \t\n\r]*$")


def durwordsen(arg):
    durWordsMatch = durwordsenPattern.match(arg)
    if not durWordsMatch or len(arg.strip()) == 0:
        raise FunctionArgType(1, "durwordsen lexical error")
    try:
        dur = 'P'
        durWordsMatchGroups = durWordsMatch.groups()
        for groupIndex, groupSuffix in ((1,"Y"), (65,"M"), (129, "D")):
            groupPart = durWordsMatchGroups[groupIndex]
            if groupPart and not durwordZeroNoPattern.match(groupPart):
                if groupPart.isnumeric():
                    dur += groupPart + groupSuffix
                else:
                    dur += str(text2num(commaAndPattern.sub(" ", groupPart.strip().lower()))) + groupSuffix
        return dur if len(dur) > 1 else "P0D" # must have at least one number and designator
    except (NumberException, TypeError, ValueError) as ex:
        raise FunctionArgType(0, str(ex))

ballotboxPattern = re_compile(r"^[ \t\n\r]*([\u2610-\u2612])[ \t\n\r]*$")

def boolballotbox(arg):
    ballotboxMatch = ballotboxPattern.match(arg)
    if not ballotboxMatch or len(arg.strip()) == 0:
        raise FunctionArgType(1, "boolballotboxMatch lexical error")
    try:
        return ("false", "true", "true")[ord(ballotboxMatch.group(1)[0]) - ord('\u2610')]
    except (IndexException, TypeError) as ex:
        raise FunctionArgType(0, str(ex))

def yesnoballotbox(arg):
    ballotboxMatch = ballotboxPattern.match(arg)
    if not ballotboxMatch or len(arg.strip()) == 0:
        raise FunctionArgType(1, "yesnoballotboxMatch lexical error")
    try:
        return ("No", "Yes", "Yes")[ord(ballotboxMatch.group(1)[0]) - ord('\u2610')]
    except (IndexException, TypeError) as ex:
        raise FunctionArgType(0, str(ex))

def codeIndex(match):
    groups = match.groups()
    for grpNbr in range(1, len(groups)):
        if groups[grpNbr] is not None:
            # ith group matched, get pattern's code position
            parenNbr = 0 # outer paren is for the enclosure of all the groups
            for codeIndex, patternForCode in enumerate(match.re.pattern.split(")|")):
                inCharacterClass = False
                for c in patternForCode:
                    if c == "[":
                        inCharacterClass = True
                    elif c == "]":
                        inCharacterClass = False
                    if c == "(" and not inCharacterClass:
                        if parenNbr == grpNbr:
                            return codeIndex # this is the code which matched
                        parenNbr += 1
    return None # no match


countryNamePattern = re_compile(r"^[ \t\n\r]*(([Aa]fghanistan)|([ÅåAa]land\s+[Ii]slands)|([Aa]lbania)|([Aa]lgeria)|([Aa]merican\s+[Ss]amoa)|([Aa]ndorra)|([Aa]ngola)|([Aa]nguilla)|([Aa]ntarctica)|([Aa]ntigua\s+[Aa]nd\s+[Bb]arbuda)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Aa]rgentina)|([Aa]rmenia)|([Aa]ruba)|([Aa]ustralia)|([Aa]ustria)|([Aa]zerbaijan)|([Bb]ahamas)|([Bb]ahrain)|([Bb]angladesh)|([Bb]arbados)|([Bb]elarus)|([Bb]elgium)|([Bb]elize)|([Bb]enin)|([Bb]ermuda)|([Bb]hutan)|([Bb]olivia)|([Bb]onaire,\s+[Ss]int\s+[Ee]ustatius\s+[Aa]nd\s+[Ss]aba)|([Bb]osnia\s+[Aa]nd\s+[Hh]erzegovina)|([Bb]otswana)|([Bb]ouvet\s+[Ii]sland)|((([Tt]he\s+)?[Ff]ederative\s+[Rr]epublic\s+[Oo]f\s+)?[Bb]ra[sz]il)|([Bb]ritish\s+[Ii]ndian\s+[Oo]cean\s+[Tt]erritory)|([Bb]runei\s+[Dd]arussalam)|([Bb]ulgaria)|([Bb]urkina\s+[Ff]aso)|([Bb]urundi)|([Cc]abo\s+[Vv]erde)|([Cc]ambodia)|([Cc]ameroon)|([Cc]anada)|([Cc]ayman\s+[Ii]slands)|([Cc]entral\s+[Aa]frican\s+[Rr]epublic)|([Cc]had)|([Cc]hile)|((([Tt]he\s+)?[Pp]eople['’]?s\s+[Rr]epublic\s+[Oo]f\s+)?[Cc]hina)|([Cc]hristmas\s+[Ii]sland)|([Cc]ocos\s+(.[Kk]eeling.\s+)?[Ii]slands)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Cc]olombia)|([Cc]omoros)|([Dd]emocratic\s+[Rr]epublic\s+[Oo]f\s+([Tt]he\s+)?[Cc]ongo)|([Cc]ongo)|([Cc]ook\s+[Ii]slands)|([Cc]osta\s+[Rr]ica)|([Cc][ôo]te\s+d['’][Ii]voire)|([Cc]roatia)|([Cc]uba)|([Cc]ura[çc]ao)|([Cc]yprus)|([Cc]zechia|[Cc]zech\s+[Rr]epublic)|((([Tt]he\s+)?[Kk]ingdom\s+[Oo]f\s+)?[Dd]enmark)|([Dd]jibouti)|([Dd]ominica)|([Dd]ominican\s+[Rr]epublic)|([Ee]cuador)|([Ee]gypt)|([Ee]l\s+[Ss]alvador)|([Ee]quatorial\s+[Gg]uinea)|([Ee]ritrea)|([Ee]stonia)|([Ee]swatini)|([Ee]thiopia)|(([Tt]he\s+)?[Ff]alkland\s+[Ii]slands(\s+.[Mm]alvinas.)?|([Ii]slas\s+)?[Mm]alvinas(\s+[Ii]slands)?)|([Ff]aroe\s+[Ii]slands)|([Ff]iji)|([Ff]inland)|([Ff]rance)|([Ff]rench\s+[Gg]uiana)|([Ff]rench\s+[Pp]olynesia)|([Ff]rench\s+[Ss]outhern\s+[Tt]erritories)|([Gg]abon)|([Gg]ambia)|([Gg]eorgia)|([Gg]ermany)|([Gg]hana)|([Gg]ibraltar)|([Gg]reece)|([Gg]reenland)|([Gg]renada)|([Gg]uadeloupe)|([Gg]uam)|([Gg]uatemala)|([Gg]uernsey)|([Gg]uinea)|([Gg]uinea-[Bb]issau)|([Gg]uyana)|([Hh]aiti)|([Hh]eard\s+[Ii]sland\s+[Aa]nd\s+[Mm]cDonald\s+[Ii]slands)|(([Tt]he\s+)?[Hh]oly\s+[Ss]ee|[Vv]atican\s+[Cc]ity)|([Hh]onduras)|([Hh]ong\s+[Kk]ong)|([Hh]ungary)|([Ii]celand)|([Ii]ndia)|([Ii]ndonesia)|((([Tt]he\s+)?[Ii]slamic\s+[Rr]epublic\s+of\s+)?[Ii]ran)|([Ii]raq)|([Ii]reland)|([Ii]sle\s+[Oo]f\s+[Mm]an)|([Ii]srael)|([Ii]taly)|([Jj]amaica)|([Jj]apan)|([Jj]ersey)|([Jj]ordan)|([Kk]azakhstan)|([Kk]enya)|([Kk]iribati)|(([Nn]orth|[Dd]emocratic\s+[Pp]eople['’]?s\s+[Rr]epublic\s+[Oo]f)\s+[Kk]orea)|(([Ss]outh|[Rr]epublic\s+[Oo]f)\s+[Kk]orea)|([Kk]uwait)|([Kk]yrgyzstan)|([Ll]ao\s+[Pp]eople['’]?s\s+[Dd]emocratic\s+[Rr]epublic)|([Ll]atvia)|([Ll]ebanon)|([Ll]esotho)|([Ll]iberia)|((([Tt]he\s+)?[Ss]tate\s+[Oo]f\s+)?[Ll]ibya)|([Ll]iechtenstein)|([Ll]ithuania)|((([Tt]he\s+)?([Gg]rand\s+)?[Dd]uchy\s+[Oo]f\s+)?[Ll]uxembourg)|([Mm]aca[ou])|([Mm]adagascar)|([Mm]alawi)|([Mm]alaysia)|([Mm]aldives)|([Mm]ali)|([Mm]alta)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+([Tt]he\s+)?)?[Mm]arshall\s+[Ii]slands)|([Mm]artinique)|([Mm]auritania)|([Mm]auritius)|([Mm]ayotte)|([Mm]exico|[Uu]nited\s+[Mm]exican\s+[Ss]tates)|((([Tt]he\s+)?[Ff]ederated\s+[Ss]tates\s+[Oo]f\s+)?[Mm]icronesia)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Mm]oldova)|([Mm]onaco)|([Mm]ongolia)|([Mm]ontenegro)|([Mm]ontserrat)|([Mm]orocco)|([Mm]ozambique)|([Mm]yanmar)|([Nn]amibia)|([Nn]auru)|([Nn]epal)|([Nn]etherlands)|([Nn]ew\s+[Cc]aledonia)|([Nn]ew\s+[Zz]ealand)|([Nn]icaragua)|([Nn]iger)|([Nn]igeria)|([Nn]iue)|([Nn]orfolk\s+[Ii]sland)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Nn]orth\s+[Mm]acedonia)|([Nn]orthern\s+[Mm]ariana\s+[Ii]slands)|([Nn]orway)|([Oo]man)|([Pp]akistan)|([Pp]alau)|((([Tt]he\s+)?[Ss]tate\s+[Oo]f\s+)?[Pp]alestine)|([Pp]anama)|([Pp]apua\s+[Nn]ew\s+[Gg]uinea)|([Pp]araguay)|([Pp]eru)|([Pp]hilippines)|([Pp]itcairn)|([Pp]oland)|([Pp]ortugal)|([Pp]uerto\s+[Rr]ico)|([Qq]atar)|([Rr][ée]union)|([Rr]omania)|([Rr]ussian\s+[Ff]ederation)|([Rr]wanda)|([Ss]aint\s+[Bb]arth[ée]lemy)|([Ss]aint\s+[Hh]elena(,\s+[Aa]scension,?\s+[Aa]nd\s+[Tt]ristan(\s+[Dd]a\s+[Cc]unha)?)?)|([Ss]aint\s+[Kk]itts\s+[Aa]nd\s+[Nn]evis)|([Ss]aint\s+[Ll]ucia)|([Ss]aint\s+[Mm]artin)|([Ss]aint\s+[Pp]ierre\s+[Aa]nd\s+[Mm]iquelon)|([Ss]aint\s+[Vv]incent(\s+[Aa]nd\s+([Tt]he\s+)?[Gg]renadines)?)|([Ss]amoa)|([Ss]an\s+[Mm]arino)|([Ss]ao\s+[Tt]ome\s+[Aa]nd\s+[Pp]rincipe)|([Ss]audi\s+[Aa]rabia)|([Ss]enegal)|([Ss]erbia)|([Ss]eychelles)|([Ss]ierra\s+[Ll]eone)|([Ss]ingapore)|([Ss]int\s+[Mm]aarten)|([Ss]lovakia)|([Ss]lovenia)|([Ss]olomon\s+[Ii]slands)|([Ss]omalia)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Ss]outh\s+[Aa]frica)|([Ss]outh\s+[Gg]eorgia\s+[Aa]nd\s+([Tt]he\s+)?[Ss]outh\s+[Ss]andwich\s+[Ii]slands)|([Ss]outh\s+[Ss]udan)|((([Tt]he\s+)?[Kk]ingdom\s+[Oo]f\s+)?[Ss]pain|[Ee]spa[ñn]a)|([Ss]ri\s+[Ll]anka)|([Ss]udan)|([Ss]uriname)|([Ss]valbard\s+[Aa]nd\s+[Jj]an\s+[Mm]ayen)|([Ss]weden)|([Ss]witzerland)|([Ss]yria(n\s+[Aa]rab\s+[Rr]epublic)?)|([Tt]aiwan(,?\s+[Pp]rovince\s+[Oo]f\s+[Cc]hina)?)|([Tt]ajikistan)|((([Tt]he\s+)?[Uu]nited\s+[Rr]epublic\s+[Oo]f\s+)?[Tt]anzania)|([Tt]hailand)|([Tt]imor-[Ll]este)|([Tt]ogo)|([Tt]okelau)|([Tt]onga)|([Tt]rinidad\s+[Aa]nd\s+[Tt]obago)|([Tt]unisia)|([Tt][uü]rk(ey|iye))|([Tt]urkmenistan)|([Tt]urks\s+[Aa]nd\s+[Cc]aicos\s+[Ii]slands)|([Tt]uvalu)|([Uu]ganda)|([Uu]kraine)|(UAE|[Uu]nited\s+[Aa]rab\s+[Ee]mirates)|(U[.]?K[.]?|[Bb]ritain|[Gg]reat\s+[BBb]ritain|[Uu]nited\s+[Kk]ingdom(\s+[Oo]f\s+[Gg]reat\s+[Bb]ritain\s+[Aa]nd\s+[Nn]orthern\s+[Ii]reland)?|[Ee]ngland(\s+[Aa]nd\s+[Ww]ales)?)|([Uu]nited\s+[Ss]tates\s+[Mm]inor\s+[Oo]utlying\s+[Ii]slands)|(U[.]?S[.]?A[.]?|[Uu]nited\s+[Ss]tates(\s+[Oo]f\s+[Aa]merica)?)|([Uu]ruguay)|([Uu]zbekistan)|([Vv]anuatu)|([Vv]enezuela)|([Vv]iet\s+[Nn]am)|([Bb]ritish\s+[Vv]irgin\s+[Ii]slands)|([Uu][.]?[Ss][.]?\s+[Vv]irgin\s+[Ii]slands)|([Ww]allis\s+[Aa]nd\s+[Ff]utuna)|([Ww]estern\s+[Ss]ahara*)|([Yy]emen)|([Zz]ambia)|([Zz]imbabwe))[ \t\n\r]*$")

countryCodes = ("AF", "AX", "AL", "DZ", "AS", "AD", "AO", "AI", "AQ", "AG", "AR", "AM", "AW", "AU", "AT", "AZ", "BS", "BH", "BD", "BB", "BY", "BE", "BZ", "BJ", "BM", "BT", "BO", "BQ", "BA", "BW", "BV", "BR", "IO", "BN", "BG", "BF", "BI", "CV", "KH", "CM", "CA", "KY", "CF", "TD", "CL", "CN", "CX", "CC", "CO", "KM", "CD", "CG", "CK", "CR", "CI", "HR", "CU", "CW", "CY", "CZ", "DK", "DJ", "DM", "DO", "EC", "EG", "SV", "GQ", "ER", "EE", "SZ", "ET", "FK", "FO", "FJ", "FI", "FR", "GF", "PF", "TF", "GA", "GM", "GE", "DE", "GH", "GI", "GR", "GL", "GD", "GP", "GU", "GT", "GG", "GN", "GW", "GY", "HT", "HM", "VA", "HN", "HK", "HU", "IS", "IN", "ID", "IR", "IQ", "IE", "IM", "IL", "IT", "JM", "JP", "JE", "JO", "KZ", "KE", "KI", "KP", "KR", "KW", "KG", "LA", "LV", "LB", "LS", "LR", "LY", "LI", "LT", "LU", "MO", "MG", "MW", "MY", "MV", "ML", "MT", "MH", "MQ", "MR", "MU", "YT", "MX", "FM", "MD", "MC", "MN", "ME", "MS", "MA", "MZ", "MM", "NA", "NR", "NP", "NL", "NC", "NZ", "NI", "NE", "NG", "NU", "NF", "MK", "MP", "NO", "OM", "PK", "PW", "PS", "PA", "PG", "PY", "PE", "PH", "PN", "PL", "PT", "PR", "QA", "RE", "RO", "RU", "RW", "BL", "SH", "KN", "LC", "MF", "PM", "VC", "WS", "SM", "ST", "SA", "SN", "RS", "SC", "SL", "SG", "SX", "SK", "SI", "SB", "SO", "ZA", "GS", "SS", "ES", "LK", "SD", "SR", "SJ", "SE", "CH", "SY", "TW", "TJ", "TZ", "TH", "TL", "TG", "TK", "TO", "TT", "TN", "TR", "TM", "TC", "TV", "UG", "UA", "AE", "GB", "UM", "US", "UY", "UZ", "VU", "VE", "VN", "VG", "VI", "WF", "EH", "YE", "ZM", "ZW")

def countrynameen(arg):
    countryMatch = countryNamePattern.match(arg)
    if not countryMatch or len(arg.strip()) == 0:
        raise FunctionArgType(1, "countrynameen lexical error")
    try:
        return countryCodes[codeIndex(countryMatch)]
    except (IndexException, TypeError) as ex:
        raise FunctionArgType(0, str(ex))

stateProvNamePattern = re_compile(r"^[ \t\n\r]*(([Aa]labama)|([Aa]laska)|([Aa]rizona)|([Aa]rkansas)|([Cc]alifornia)|([Cc]olorado)|([Cc]onnecticut)|([Dd]elaware)|([Ff]lorida)|([Gg]eorgia)|([Hh]awaii)|([Ii]daho)|([Ii]llinois)|([Ii]ndiana)|([Ii]owa)|([Kk]ansas)|([Kk]entucky)|([Ll]ouisiana)|([Mm]aine)|([Mm]aryland)|([Mm]assachusetts)|([Mm]ichigan)|([Mm]innesota)|([Mm]ississippi)|([Mm]issouri)|([Mm]ontana)|([Nn]ebraska)|([Nn]evada)|([Nn]ew\s+[Hh]ampshire)|([Nn]ew\s+[Jj]ersey)|([Nn]ew\s+[Mm]exico)|([Nn]ew\s+[Yy]ork)|([Nn]orth\s+[Cc]arolina)|([Nn]orth\s+[Dd]akota)|([Oo]hio)|([Oo]klahoma)|([Oo]regon)|([Pp]ennsylvania)|([Rr]hode\s+[Ii]sland)|([Ss]outh\s+[Cc]arolina)|([Ss]outh\s+[Dd]akota)|([Tt]ennessee)|([Tt]exas)|([Uu]tah)|([Vv]ermont)|([Vv]irginia)|([Ww]ashington)|([Ww]est\s+[Vv]irginia)|([Ww]isconsin)|([Ww]yoming)|([Dd]istrict\s+[Oo]f\s+[Cc]olumbia)|(([Aa]merican\s+)?[Ss]amoa)|([Gg]uam)|([Nn]orthern\s+[Mm]ariana\s+[Ii]slands)|([Pp]uerto\s+[Rr]ico)|([Uu]nited\s+[Ss]tates\s+[Mm]inor\s+[Oo]utlying\s+[Ii]slands)|([Vv]irgin\s+[Ii]slands,\s+U[.]S[.])|([Aa]lberta)|([Bb]ritish\s+[Cc]olumbia)|([Mm]anitoba)|([Nn]ew\s+[Bb]runswick)|([Nn]ewfoundland(\s+[Aa]nd\s+[Ll]abrador)?)|([Nn]ova\s+[Ss]cotia)|([Nn]orthwest\s+[Tt]erritories)|([Nn]unavut)|([Oo]ntario)|([Pp]rince\s+[Ee]dward\s+[Ii]sland)|([Qq]u[eé]bec)|([Ss]askatchewan)|([Yy]ukon))[ \t\n\r]*$")

stateProvCodes = ("AL", "AK", "AZ", "AR", "CA", "CO", "CT", "DE", "FL", "GA", "HI", "ID", "IL", "IN", "IA", "KS", "KY", "LA", "ME", "MD", "MA", "MI", "MN", "MS", "MO", "MT", "NE", "NV", "NH", "NJ", "NM", "NY", "NC", "ND", "OH", "OK", "OR", "PA", "RI", "SC", "SD", "TN", "TX", "UT", "VT", "VA", "WA", "WV", "WI", "WY", "DC", "AS", "GU", "MP", "PR", "UM", "VI", "AB", "BC", "MB", "NB", "NL", "NS", "NT", "NU", "ON", "PE", "QC", "SK", "YT")

def stateprovnameen(arg):
    stateProvMatch = stateProvNamePattern.match(arg)
    if not stateProvMatch or len(arg.strip()) == 0:
        raise FunctionArgType(1, "stateprovnameen lexical error")
    try:
        return stateProvCodes[codeIndex(stateProvMatch)]
    except (IndexException, TypeError) as ex:
        raise FunctionArgType(0, str(ex))

edgarProvCountryNamePattern = re_compile(r"^[ \t\n\r]*(([Aa]merican\s+[Ss]amoa)|([Gg]uam)|([Pp]uerto\s+[Rr]ico)|([Uu]nited\s+[Ss]tates\s+[Mm]inor\s+[Oo]utlying\s+[Ii]slands)|([Vv]irgin\s+[Ii]slands,\s+U[.]?S[.]?|U([.]\s*)?S[.]?\s+[Vv]irgin\s+[Ii]slands)|([Aa]lberta(,?\s+[Cc]anada)?)|([Bb]ritish\s+[Cc]olumbia(,?\s+[Cc]anada)?)|([Mm]anitoba(,?\s+[Cc]anada)?)|([Nn]ew\s+[Bb]runswick(,?\s+[Cc]anada)?)|([Nn]ewfoundland(\s+[Aa]nd\s+[Ll]abrador)?(,?\s+[Cc]anada)?)|([Nn]ova\s+[Ss]cotia(,?\s+[Cc]anada)?)|([Oo]ntario(,?\s+[Cc]anada)?)|([Pp]rince\s+[Ee]dward\s+[Ii]sland(,?\s+[Cc]anada)?)|([Qq]u[eé]bec(,?\s+[Cc]anada)?)|([Ss]askatchewan(,?\s+[Cc]anada)?)|([Yy]ukon(,?\s+[Cc]anada)?)|([Aa]fghanistan)|([ÅåAa]land\s+[Ii]slands)|([Aa]lbania)|([Aa]lgeria)|([Aa]merican\s+[Ss]amoa)|([Aa]ndorra)|([Aa]ngola)|([Aa]nguilla)|([Aa]ntarctica)|([Aa]ntigua\s+[Aa]nd\s+[Bb]arbuda)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Aa]rgentina)|([Aa]rmenia)|([Aa]ruba)|([Aa]ustralia)|([Aa]ustria)|([Aa]zerbaijan)|([Bb]ahamas)|([Bb]ahrain)|([Bb]angladesh)|([Bb]arbados)|([Bb]elarus)|([Bb]elgium)|([Bb]elize)|([Bb]enin)|([Bb]ermuda)|([Bb]hutan)|([Bb]olivia)|([Bb]osnia\s+[Aa]nd\s+[Hh]erzegovina)|([Bb]otswana)|([Bb]ouvet\s+[Ii]sland)|((([Tt]he\s+)?[Ff]ederative\s+[Rr]epublic\s+[Oo]f\s+)?[Bb]ra[sz]il)|([Bb]ritish\s+[Ii]ndian\s+[Oo]cean\s+[Tt]erritory)|([Bb]runei\s+[Dd]arussalam)|([Bb]ulgaria)|([Bb]urkina\s+[Ff]aso)|([Bb]urundi)|([Cc]a(bo|pe)\s+[Vv]erde)|([Cc]ambodia)|([Cc]ameroon)|([Cc]anada)|([Cc]ayman\s+[Ii]slands)|([Cc]entral\s+[Aa]frican\s+[Rr]epublic)|([Cc]had)|([Cc]hile)|((([Tt]he\s+)?[Pp]eople['’]?s\s+[Rr]epublic\s+[Oo]f\s+)?[Cc]hina)|([Cc]hristmas\s+[Ii]sland)|([Cc]ocos\s+(.[Kk]eeling.\s+)?[Ii]slands)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Cc]olombia)|([Cc]omoros)|([Dd]emocratic\s+[Rr]epublic\s+[Oo]f\s+([Tt]he\s+)?[Cc]ongo)|([Cc]ongo)|([Cc]ook\s+[Ii]slands)|([Cc]osta\s+[Rr]ica)|([Cc][ôo]te\s+d['’][Ii]voire)|([Cc]roatia)|([Cc]uba)|([Cc]yprus)|([Cc]zechia|[Cc]zech\s+[Rr]epublic)|((([Tt]he\s+)?[Kk]ingdom\s+[Oo]f\s+)?[Dd]enmark)|([Dd]jibouti)|([Dd]ominica)|([Dd]ominican\s+[Rr]epublic)|([Ee]cuador)|([Ee]gypt)|([Ee]l\s+[Ss]alvador)|([Ee]quatorial\s+[Gg]uinea)|([Ee]ritrea)|([Ee]stonia)|([Ee]thiopia)|(([Tt]he\s+)?[Ff]alkland\s+[Ii]slands(\s+.[Mm]alvinas.)?|([Ii]slas\s+)?[Mm]alvinas(\s+[Ii]slands)?)|([Ff]aroe\s+[Ii]slands)|([Ff]iji)|([Ff]inland)|([Ff]rance)|([Ff]rench\s+[Gg]uiana)|([Ff]rench\s+[Pp]olynesia)|([Ff]rench\s+[Ss]outhern\s+[Tt]erritories)|([Gg]abon)|([Gg]ambia)|([Gg]eorgia)|([Gg]ermany)|([Gg]hana)|([Gg]ibraltar)|([Gg]reece)|([Gg]reenland)|([Gg]renada)|([Gg]uadeloupe)|([Gg]uatemala)|([Gg]uernsey)|([Gg]uinea)|([Gg]uinea-[Bb]issau)|([Gg]uyana)|([Hh]aiti)|([Hh]eard\s+[Ii]sland\s+[Aa]nd\s+[Mm]cDonald\s+[Ii]slands)|(([Tt]he\s+)?[Hh]oly\s+[Ss]ee|[Vv]atican\s+[Cc]ity)|([Hh]onduras)|([Hh]ong\s+[Kk]ong)|([Hh]ungary)|([Ii]celand)|([Ii]ndia)|([Ii]ndonesia)|((([Tt]he\s+)?[Ii]slamic\s+[Rr]epublic\s+of\s+)?[Ii]ran)|([Ii]raq)|([Ii]reland)|([Ii]sle\s+[Oo]f\s+[Mm]an)|([Ii]srael)|([Ii]taly)|([Jj]amaica)|([Jj]apan)|([Jj]ersey)|([Jj]ordan)|([Kk]azakhstan)|([Kk]enya)|([Kk]iribati)|(([Nn]orth|[Dd]emocratic\s+[Pp]eople['’]?s\s+[Rr]epublic\s+[Oo]f)\s+[Kk]orea)|(([Ss]outh|[Rr]epublic\s+[Oo]f)\s+[Kk]orea)|([Kk]uwait)|([Kk]yrgyzstan)|([Ll]ao\s+[Pp]eople['’]?s\s+[Dd]emocratic\s+[Rr]epublic)|([Ll]atvia)|([Ll]ebanon)|([Ll]esotho)|([Ll]iberia)|((([Tt]he\s+)?[Ss]tate\s+[Oo]f\s+)?[Ll]ibya)|([Ll]iechtenstein)|([Ll]ithuania)|((([Tt]he\s+)?([Gg]rand\s+)?[Dd]uchy\s+[Oo]f\s+)?[Ll]uxembourg)|([Mm]aca[ou])|([Mm]adagascar)|([Mm]alawi)|([Mm]alaysia)|([Mm]aldives)|([Mm]ali)|([Mm]alta)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+([Tt]he\s+)?)?[Mm]arshall\s+[Ii]slands)|([Mm]artinique)|([Mm]auritania)|([Mm]auritius)|([Mm]ayotte)|([Mm]exico|[Uu]nited\s+[Mm]exican\s+[Ss]tates)|((([Tt]he\s+)?[Ff]ederated\s+[Ss]tates\s+[Oo]f\s+)?[Mm]icronesia)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Mm]oldova)|([Mm]onaco)|([Mm]ongolia)|([Mm]ontenegro)|([Mm]ontserrat)|([Mm]orocco)|([Mm]ozambique)|([Mm]yanmar)|([Nn]amibia)|([Nn]auru)|([Nn]epal)|([Nn]etherlands\s+[Aa]ntilles)|([Nn]etherlands)|([Nn]ew\s+[Cc]aledonia)|([Nn]ew\s+[Zz]ealand)|([Nn]icaragua)|([Nn]iger)|([Nn]igeria)|([Nn]iue)|([Nn]orfolk\s+[Ii]sland)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?([Nn]orth\s+)?[Mm]acedonia)|([Nn]orthern\s+[Mm]ariana\s+[Ii]slands)|([Nn]orway)|([Oo]man)|([Pp]akistan)|([Pp]alau)|((([Tt]he\s+)?[Ss]tate\s+[Oo]f\s+)?[Pp]alestine)|([Pp]anama)|([Pp]apua\s+[Nn]ew\s+[Gg]uinea)|([Pp]araguay)|([Pp]eru)|([Pp]hilippines)|([Pp]itcairn)|([Pp]oland)|([Pp]ortugal)|([Qq]atar)|([Rr][ée]union)|([Rr]omania)|([Rr]ussian\s+[Ff]ederation)|([Rr]wanda)|([Ss]aint\s+[Bb]arth[ée]lemy)|([Ss]aint\s+[Hh]elena(,\s+[Aa]scension,?\s+[Aa]nd\s+[Tt]ristan(\s+[Dd]a\s+[Cc]unha)?)?)|([Ss]aint\s+[Kk]itts\s+[Aa]nd\s+[Nn]evis)|([Ss]aint\s+[Ll]ucia)|([Ss]aint\s+[Mm]artin)|([Ss]aint\s+[Pp]ierre\s+[Aa]nd\s+[Mm]iquelon)|([Ss]aint\s+[Vv]incent(\s+[Aa]nd\s+([Tt]he\s+)?[Gg]renadines)?)|([Ss]amoa)|([Ss]an\s+[Mm]arino)|([Ss]ao\s+[Tt]ome\s+[Aa]nd\s+[Pp]rincipe)|([Ss]audi\s+[Aa]rabia)|([Ss]enegal)|([Ss]erbia)|([Ss]eychelles)|([Ss]ierra\s+[Ll]eone)|([Ss]ingapore)|([Ss]lovakia)|([Ss]lovenia)|([Ss]olomon\s+[Ii]slands)|([Ss]omalia)|((([Tt]he\s+)?[Rr]epublic\s+[Oo]f\s+)?[Ss]outh\s+[Aa]frica)|([Ss]outh\s+[Gg]eorgia\s+[Aa]nd\s+([Tt]he\s+)?[Ss]outh\s+[Ss]andwich\s+[Ii]slands)|((([Tt]he\s+)?[Kk]ingdom\s+[Oo]f\s+)?[Ss]pain|[Ee]spa[ñn]a)|([Ss]ri\s+[Ll]anka)|([Ss]udan)|([Ss]uriname)|([Ss]valbard\s+[Aa]nd\s+[Jj]an\s+[Mm]ayen)|([Ss]waziland)|([Ss]weden)|([Ss]witzerland)|([Ss]yria(n\s+[Aa]rab\s+[Rr]epublic)?)|([Tt]aiwan(,?\s+[Pp]rovince\s+[Oo]f\s+[Cc]hina)?)|([Tt]ajikistan)|((([Tt]he\s+)?[Uu]nited\s+[Rr]epublic\s+[Oo]f\s+)?[Tt]anzania)|([Tt]hailand)|([Tt]imor-[Ll]este)|([Tt]ogo)|([Tt]okelau)|([Tt]onga)|([Tt]rinidad\s+[Aa]nd\s+[Tt]obago)|([Tt]unisia)|([Tt][uü]rk(ey|iye))|([Tt]urkmenistan)|([Tt]urks\s+[Aa]nd\s+[Cc]aicos\s+[Ii]slands)|([Tt]uvalu)|([Uu]ganda)|([Uu]kraine)|(UAE|[Uu]nited\s+[Aa]rab\s+[Ee]mirates)|(U[.]?K[.]?|[Bb]ritain|[Gg]reat\s+[BBb]ritain|[Uu]nited\s+[Kk]ingdom(\s+[Oo]f\s+[Gg]reat\s+[Bb]ritain\s+[Aa]nd\s+[Nn]orthern\s+[Ii]reland)?|[Ee]ngland(\s+[Aa]nd\s+[Ww]ales)?)|([Uu]nited\s+[Ss]tates\s+[Mm]inor\s+[Oo]utlying\s+[Ii]slands)|(U[.]?S[.]?A[.]?|[Uu]nited\s+[Ss]tates(\s+[Oo]f\s+[Aa]merica)?)|([Uu]ruguay)|([Uu]zbekistan)|([Vv]anuatu)|([Vv]enezuela)|([Vv]iet\s+[Nn]am)|([Bb]ritish\s+[Vv]irgin\s+[Ii]slands|[Vv]irgin\s+[Ii]slands,?\s+[Bb]ritish)|([Ww]allis\s+[Aa]nd\s+[Ff]utuna)|([Ww]estern\s+[Ss]ahara*)|([Yy]emen)|([Zz]ambia)|([Zz]imbabwe)|([Uu]nknown))[ \t\n\r]*$")

edgarProvCountryCodes = ("B5", "GU", "PR", "2J", "VI", "A0", "A1", "A2", "A3", "A4", "A5", "A6", "A7", "A8", "A9", "B0", "B2", "Y6", "B3", "B4", "B5", "B6", "B7", "1A", "B8", "B9", "C1", "1B", "1C", "C3", "C4", "1D", "C5", "C6", "C7", "C8", "1F", "C9", "D1", "G6", "D0", "D2", "D3", "1E", "B1", "D4", "D5", "D6", "D9", "E0", "X2", "E2", "E8", "E3", "E4", "Z4", "E9", "F0", "F2", "F3", "F4", "F6", "F7", "F8", "F9", "Y3", "G0", "G1", "G2", "L7", "1M", "G3", "G4", "2N", "G7", "1G", "G9", "G8", "H1", "H2", "H3", "H4", "1J", "1H", "H5", "H7", "H6", "H8", "H9", "I0", "I3", "I4", "2C", "I5", "I6", "2Q", "2M", "J0", "J1", "J3", "J4", "J5", "J6", "J8", "Y7", "J9", "S0", "K0", "K1", "K4", "X4", "K2", "K3", "K5", "K6", "K7", "K8", "K9", "L0", "L2", "Y8", "L3", "L6", "L8", "M0", "Y9", "M2", "1P", "M3", "J2", "M4", "M5", "M6", "1N", "M7", "1R", "M8", "M9", "N0", "N1", "N2", "1Q", "N4", "N5", "N6", "N7", "N8", "N9", "O0", "O1", "1T", "O2", "O3", "O4", "2P", "O5", "1K", "1S", "O9", "P0", "Z5", "P1", "P2", "P3", "E1", "T6", "P5", "P6", "P8", "P7", "1W", "Q2", "Q3", "Q4", "Q5", "Q6", "Q7", "1U", "1V", "Q8", "P4", "R0", "1Y", "1X", "R1", "R2", "R4", "R5", "R6", "R8", "R9", "S1", "S3", "S4", "S5", "1Z", "S6", "Z0", "U8", "U7", "U9", "Z1", "V0", "V1", "Y0", "S8", "S9", "T0", "T1", "Z2", "T2", "T8", "U0", "2B", "2A", "D7", "U1", "T3", "1L", "U3", "F1", "V2", "V3", "L9", "V6", "V7", "V8", "V9", "F5", "2D", "W0", "W1", "Z3", "W2", "W3", "W4", "W5", "W6", "W8", "2E", "W7", "2G", "W9", "2H", "C0", "X0", "2J", "X1", "X3", "2K", "2L", "X5", "Q1", "D8", "X8", "U5", "T7", "Y4", "Y5", "XX")

def edgarprovcountryen(arg):
    edgarProvCountryMatch = edgarProvCountryNamePattern.match(arg)
    if not edgarProvCountryMatch or len(arg.strip()) == 0:
        raise FunctionArgType(1, "stateorcountrynameen lexical error")
    try:
        return edgarProvCountryCodes[codeIndex(edgarProvCountryMatch)]
    except (IndexException, TypeError) as ex:
        raise FunctionArgType(0, str(ex))

exchNamePattern = re_compile(r"^[ \t\n\r]*((([Tt]he\s+)?[Bb][Oo][Xx]\s+[Ee]xchange(,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Cc]boe\s+[Bb][Yy][Xx]\s+[Ee]xchange(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Cc]boe\s+[Bb][Zz][Xx]\s+[Ee]xchange(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Cc]boe\s+[Cc]2\s+[Ee]xchange(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Cc]boe\s+[Ee][Dd][Gg][Aa]\s+[Ee]xchange(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Cc]boe\s+[Ee][Dd][Gg][Xx]\s+[Ee]xchange(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Cc]boe\s+[Ee]xchange(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Cc]hicago\s+[Ss]tock\s+[Ee]xchange(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Ii]nvestors\s+[Ee]xchange(,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Mm]iami\s+[Ii]nternational\s+[Ss]ecurities\s+[Ee]xchange(,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Mm]IAX\s+[Pp]EARL(,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Nn][Aa][Ss][Dd][Aa][Qq]\s+[Bb][Xx](,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Nn][Aa][Ss][Dd][Aa][Qq]\s+[Gg][Ee][Mm][Xx](,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Nn][Aa][Ss][Dd][Aa][Qq]\s+[Ii][Ss][Ee](,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Nn][Aa][Ss][Dd][Aa][Qq]\s+[Mm][Rr][Xx](,?\s+[Ll][Ll][Cc])?)|([Nn][Aa][Ss][Dd][Aa][Qq]\s+[Pp][Hh][Ll][Xx](,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Nn][Yy][Ss][Ee]|([Tt]he\s+)?[Nn]ew\s+[Yy]ork\s+[Ss]tock\s+[Ee]xchange(,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Nn][Yy][Ss][Ee]\s+[Aa]merican(,?\s+[Ll][Ll][Cc])?)|(([Tt]he\s+)?[Nn][Yy][Ss][Ee]\s+[Aa]rca(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Nn][Yy][Ss][Ee]\s+[Nn]ational(,?\s+[Ii]nc[.]?)?)|(([Tt]he\s+)?[Nn][Aa][Ss][Dd][Aa][Qq](\s+([Ss]tock|[Gg]lobal(\s+[Ss]elect)?)\s+[Mm]arket(,?\s+[Ll][Ll][Cc])?)?))[ \t\n\r]*$")

exchCodes = ("BOX", "CboeBYX", "CboeBZX", "C2", "CboeEDGA", "CboeEDGX", "CBOE", "CHX", "IEX", "MIAX", "PEARL", "BX", "GEMX", "ISE", "MRX", "Phlx", "NYSE", "NYSEAMER", "NYSEArca", "NYSENAT", "NASDAQ")

def exchnameen(arg):
    exchMatch = exchNamePattern.match(arg)
    if not exchMatch or len(arg.strip()) == 0:
        raise FunctionArgType(1, "exchnameen lexical error")
    try:
        return exchCodes[codeIndex(exchMatch)]
    except (IndexException, TypeError) as ex:
        raise FunctionArgType(0, str(ex))

entityFilerNamePattern = re_compile(r"^[ \t\n\r]*(([Ll]arge\s+[Aa]ccelerated\s+[Ff]iler)|([Aa]ccelerated\s+[Ff]iler)|([Nn]on[^\w]+[Aa]ccelerated\s+[Ff]iler))[ \t\n\r]*$")

entityFilerCodes = ("Large Accelerated Filer", "Accelerated Filer", "Non-accelerated Filer")


def entityfilercategoryen(arg):
    entityFilerMatch = entityFilerNamePattern.match(arg)
    if not entityFilerMatch or len(arg.strip()) == 0:
        raise FunctionArgType(1, "entityfilercategoryen lexical error")
    try:
        return entityFilerCodes[codeIndex(entityFilerMatch)]
    except (IndexException, TypeError) as ex:
        raise FunctionArgType(0, str(ex))

def loadSECtransforms(customTransforms, *args, **kwargs):
    ixtSEC = "http://www.sec.gov/inlineXBRL/transformation/2015-08-31"
    customTransforms.update({
        qname(ixtSEC, "ixt-sec:duryear"): duryear,
        qname(ixtSEC, "ixt-sec:durmonth"): durmonth,
        qname(ixtSEC, "ixt-sec:durweek"): durweek,
        qname(ixtSEC, "ixt-sec:durday"): durday,
        qname(ixtSEC, "ixt-sec:durhour"): durhour,
        qname(ixtSEC, "ixt-sec:datequarterend"): datequarterend,
        qname(ixtSEC, "ixt-sec:numinf"): numinf,
        qname(ixtSEC, "ixt-sec:numneginf"): numneginf,
        qname(ixtSEC, "ixt-sec:numnan"): numnan,
        qname(ixtSEC, "ixt-sec:numwordsen"): numwordsen,
        qname(ixtSEC, "ixt-sec:durwordsen"): durwordsen,
        qname(ixtSEC, "ixt-sec:boolballotbox"): boolballotbox,
        qname(ixtSEC, "ixt-sec:yesnoballotbox"): yesnoballotbox,
        qname(ixtSEC, "ixt-sec:countrynameen"): countrynameen,
        qname(ixtSEC, "ixt-sec:stateprovnameen"): stateprovnameen,
        qname(ixtSEC, "ixt-sec:edgarprovcountryen"): edgarprovcountryen,
        qname(ixtSEC, "ixt-sec:exchnameen"): exchnameen,
        qname(ixtSEC, "ixt-sec:entityfilercategoryen"): entityfilercategoryen
    })

__pluginInfo__ = {
    'name': 'SEC Inline Transforms',
    'version': '19.2', # SEC version
    'description': "This plug-in adds custom transforms SEC inline filing with durations.  ",
    'license': 'Apache-2',
    'author': f'SEC employees (integrated by {authorLabel})',
    'copyright': f'{copyrightLabel} \nUtilizes text2num.py (c) 2008 Greg Hewgill',
    # classes of mount points (required)
    'ModelManager.LoadCustomTransforms': loadSECtransforms,
}



'''
Created by staff of the U.S. Securities and Exchange Commission.
Data and content created by government employees within the scope of their employment are not subject
to domestic copyright protection. 17 U.S.C. 105.

To run
   arelleCmdLine --plugin validate/EFM/tools/CheckTxmyRefs --disclosureSystem efm-pragmatic --report-file ../myReportFile.csv

   if no --report-file the log entryies are printed to standard output stream

pass 1: for each edgartaxonomy family, for each version,
        load each href
        check that it's in the standard taxonomies dict inferred from edgartaxonomies
        check that each referenced URL from the href is also an href in edgartaxonomies
        check that any ref and doc LBs in same cache dir are in TaxonomyAddonManager files

pass 2: for each file found in an local-root directory by entries in fishInDirsForMissing
        for pattern (such as fishing for files -sub-)
        check if the file is loadable and if so if there are referenced files not in edgartaxonomies

'''
import os, regex as re
from collections import defaultdict
from lxml import etree
from arelle.Version import authorLabel, copyrightLabel

fishInDirsForMissing = {
    "stx": {
        "url-root": "https://xbrl.sec.gov",
        # "local-root": r"C:/gns/w/git-osd/stx/xbrl.sec.gov",
        "local-root": r"C:/gns/w/git-group2/sec-arelle/xbrl_taxonomies/xbrl_taxonomies/resources/System/cache/https/xbrl.sec.gov",
        "entrypoint-pattern": "https?://xbrl.sec.gov/[^/]+/2023[^/]*/[^/]+(-sub-)[^.]+\.xsd"
        }
    # not sure what we load for US-GAAP and IFRS to determine needed ref and doc linkbases
    }


def utilityRun(cntlr, options, *args, **kwargs):
    logLines = []
    def log(line):
        if options.reportFile:
            logLines.append(line)
        else:
            print(line)
    disclosureSystem = cntlr.modelManager.disclosureSystem
    edTxmyTree = etree.parse(disclosureSystem.standardTaxonomiesUrl)
    edTxmyFamilies = defaultdict(dict) # by Family, Version, set of Hrefs
    edTxmyHrefs = set()
    for locElt in edTxmyTree.iter("Loc"):
        family = locElt.find("Family").text
        version = locElt.find("Version").text
        href = locElt.find("Href").text
        txmyFamily = edTxmyFamilies[family]
        txmyFamily.setdefault(version, set()).add(href)
        edTxmyHrefs.add(href)

    txmyAddOnTree = etree.parse(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__)))), "EdgarRenderer", "resources", "TaxonomyAddonManager.xml"))
    txmyAddons = defaultdict(dict)
    for txmyElt in txmyAddOnTree.iter("TaxonomyAddon"):
        txmyAddon = txmyAddons[txmyElt.find("Taxonomy").text]
        defsElt = txmyElt.find("DefinitionFiles")
        if defsElt is not None:
            for elt in defsElt.iter("string"):
                txmyAddon.setdefault("doc", set()).add(elt.text)
        refsElt = txmyElt.find("ReferenceFiles")
        if refsElt is not None:
            for elt in refsElt.iter("string"):
                txmyAddon.setdefault("ref", set()).add(elt.text)

    # check each edgartaxonomy family
    for txmyFamily in edTxmyFamilies:
        for familyVer in edTxmyFamilies[txmyFamily]:
            for hRef in edTxmyFamilies[txmyFamily][familyVer]:
                # check if it loads
                modelXbrl = cntlr.modelManager.load(hRef)
                if modelXbrl is not None:
                    log(f"family {txmyFamily} ver {familyVer} href {hRef} loaded")
                    if not hRef in disclosureSystem.standardTaxonomiesDict:
                        log(f"family {txmyFamily} ver {familyVer} href {hRef} not in edgartaxonomies")
                    for refUrl in modelXbrl.urlDocs:
                        if refUrl not in edTxmyHrefs and refUrl != hRef:
                            log(f"ref taxonomy also missing from edgartaxonomies: ref {refUrl}, loaded from {hRef}")
                    if not re.match(".*([-_](lab|ref|pre|cal|def|sub)(-\n\n\n\n)?\.xsd)", hRef):
                        cachePath = os.path.dirname(cntlr.webCache.getfilename(hRef))
                        docRefs = defaultdict(set)
                        for root, dir, files in os.walk(cachePath):
                            for file in files:
                                m = re.match(".*(doc|ref)\.(xsd|xml)", file)
                                if m:
                                    docRefs[m.group(1)].add(file)
                        addOn = txmyAddons.get(os.path.basename(hRef))
                        if addOn:
                            if docRefs != addOn:
                                for _type in ("doc", "ref"):
                                    missing = ", ".join(set(docRefs.get(_type, {}.values())) - set(addOn.get(_type, {}.values())))
                                    extra = ", ".join(set(addOn.get(_type, {}.values())) - set(docRefs.get(_type, {}.values())))
                                    log(f"family {txmyFamily} ver {familyVer} href {hRef}, addon extra {_type}s: {extra}, addon missing {_type}s: {missing}")
                        elif "-sub-" not in hRef and not hRef.endswith(".xml"):
                            log(f"family {txmyFamily} ver {familyVer} href {hRef} not in TaxonomyAddons")
                else:
                    log(f"family {txmyFamily} ver {familyVer} href {hRef} DID NOT LOAD")
    missingHrefs = set(disclosureSystem.standardTaxonomiesDict) - edTxmyHrefs
    extraHrefs = edTxmyHrefs - set(disclosureSystem.standardTaxonomiesDict)
    if missingHrefs:
        log(f"disclosure system missing {', '.join(sorted(missingHrefs))}")
    if extraHrefs:
        log(f"disclosure system missing {', '.join(sorted(extraHrefs))}")
    # fish in unzipped archive for file patterns
    for repo in fishInDirsForMissing:
        fileRoot = fishInDirsForMissing[repo]["local-root"]
        urlRoot = fishInDirsForMissing[repo]["url-root"]
        urlPattern = re.compile(fishInDirsForMissing[repo]["entrypoint-pattern"])
        for root, dir, files in os.walk(fileRoot):
            for file in files:
                fullFilePath = os.path.join(root, file)
                relativePath = fullFilePath[len(fileRoot)+1:]
                hRef = os.path.join(urlRoot, relativePath)
                if urlPattern.match(hRef):
                    if hRef not in edTxmyHrefs:
                        # check if it loads
                        modelXbrl = cntlr.modelManager.load(hRef)
                        if modelXbrl is not None:
                            log(f"missing from edgartaxonomies: href {hRef}, loaded")
                            for refUrl in modelXbrl.urlDocs:
                                if refUrl not in edTxmyHrefs and refUrl != hRef:
                                    log(f"ref taxonomy also missing from edgartaxonomies: ref {refUrl}, loaded from {hRef}")
                            if not re.match(".*([-_](lab|ref|pre|cal|def|sub)(-\n\n\n\n)?\.xsd)", hRef):
                                cachePath = os.path.dirname(cntlr.webCache.getfilename(hRef))
                                docRefs = defaultdict(set)
                                for root, dir, files in os.walk(cachePath):
                                    for file in files:
                                        m = re.match(".*(doc|ref)\.(xsd|xml)", file)
                                        if m:
                                            docRefs[m.group(1)].add(file)
                                addOn = txmyAddons.get(os.path.basename(hRef))
                                if not addOn:
                                    log(f"ref {hRef} not in TaxonomyAddons")
                                else:
                                    if docRefs != addOn:
                                        for _type in ("doc", "ref"):
                                            missing = ", ".join(set(docRefs.get(_type, {}.values())) - set(addOn.get(_type, {}.values())))
                                            extra = ", ".join(set(addOn.get(_type, {}.values())) - set(docRefs.get(_type, {}.values())))
                                            log(f"family {txmyFamily} ver {familyVer} href {hRef}, addon extra {_type}s: {extra}, addon missing {_type}s: {missing}")
                        else:
                            log(f"missing from edgartaxonomies: {hRef} DID NOT LOAD")
    if options.reportFile:
        with open(options.reportFile, "w", encoding="utf-8") as fh:
            fh.write("\n".join(logLines))

def commandLineOptionExtender(parser, *args, **kwargs):
    # extend command line options to store to database
    parser.add_option("--report-file",
                      action="store",
                      dest="reportFile",
                      help=_("Report file on taxonomy refs."))

def dislosureSystemTypes(disclosureSystem, *args, **kwargs):
    # return ((disclosure system name, variable name), ...)
    return (("EFM", "EFMplugin"),)

def disclosureSystemConfigURL(disclosureSystem, *args, **kwargs):
    return os.path.join(os.path.dirname(os.path.dirname(__file__)), "config.xml")


__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Check Taxonomy Refs',
    'version': '1.0',
    'description': '''
Check edgartaxonomies.xml and taxonomyAddonManager.xml
for completeness against supported year entry points.''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrCmdLine.Options': commandLineOptionExtender,
    'CntlrCmdLine.Utility.Run': utilityRun,
    'DisclosureSystem.Types': dislosureSystemTypes,
    'DisclosureSystem.ConfigURL': disclosureSystemConfigURL,
}



'''
pluginPackages test case

See COPYRIGHT.md for copyright information.
'''
# this module would raise system error due to PEP 366 after python 3.4.3
from . import importTestImported11
from arelle.Version import authorLabel, copyrightLabel

def foo():
    print ("imported unpackaged plug-in relative imported 1")

__pluginInfo__ = {
    'name': 'Unpackaged Relative Import 1',
    'version': '0.9',
    'description': "This is a unpackaged relative imported plugin.",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'Import.Unpackaged.Entry6': foo,
    # imported plugins
}



'''
Sample custom functions plugin for formula custom functions

See COPYRIGHT.md for copyright information.
'''
from arelle import XbrlUtil
from arelle.formula import XPathContext
from arelle.ModelValue import qname
from arelle.ModelInstanceObject import ModelDimensionValue
from arelle.Version import authorLabel, copyrightLabel
from decimal import Decimal

# custom function for test case 22015 v01, same as in FunctionCustom.py
def  test_22015v01_my_fn_PDxEV(xc, p, contextItem, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    PDseq = args[0] if isinstance(args[0],(list,tuple)) else (args[0],)
    EVseq = args[1] if isinstance(args[1],(list,tuple)) else (args[1],)
    dimQname = qname("{http://www.example.com/wgt-avg}ExposuresDimension")
    PDxEV = []
    for pd in PDseq:
        if pd.context is not None:
            pdDim = pd.context.dimValue(dimQname)
            for ev in EVseq:
                if ev.context is not None:
                    evDim = ev.context.dimValue(dimQname)
                    if pdDim is not None and isinstance(pdDim,ModelDimensionValue):
                        dimEqual =  pdDim.isEqualTo(evDim, equalMode=XbrlUtil.S_EQUAL2)
                    elif evDim is not None and isinstance(evDim,ModelDimensionValue):
                        dimEqual =  evDim.isEqualTo(pdDim, equalMode=XbrlUtil.S_EQUAL2)
                    else:
                        dimEqual = (pdDim == evDim)
                    if dimEqual:
                        PDxEV.append(pd.xValue * ev.xValue)
                        # type promotion required
                        if isinstance(pdX,Decimal) and isinstance(evX,float):
                            pdX = float(pdX)
                        elif isinstance(evX,Decimal) and isinstance(pdX,float):
                            pdX = float(evX)
                        PDxEV.append(pdX * evX)
                        break
    return PDxEV

# returns dict of function QName and method executing the function
def customFunctions():
    return {
        # sample function included for formula tests
        qname("{http://www.example.com/wgt-avg/function}my-fn:PDxEV"): test_22015v01_my_fn_PDxEV
    }

__pluginInfo__ = {
    'name': 'Custom Formula Functions (example)',
    'version': '1.0',
    'description': "This plug-in adds a custom function implemented by a plug-in.  ",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'Formula.CustomFunctions': customFunctions,
}



from pathlib import PurePath
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig

TIMING = {f'esef_conformance_suite_2021/esef_conformance_suite_2021/tests/inline_xbrl/{k}': v for k, v in {
    'RTS_Annex_III_Par_3_G3-1-3/index.xml': 4.901,
    'RTS_Art_6_a/index.xml': 6.769,
    'G2-5-4_2/index.xml': 7.269,
    'RTS_Annex_IV_Par_14_G2-5-1/index.xml': 7.325,
    'RTS_Annex_IV_Par_6/index.xml': 7.347,
    'RTS_Annex_IV_Par_4_1/index.xml': 7.356,
    'RTS_Annex_II_Par_1_RTS_Annex_IV_par_7/index.xml': 7.37,
    'G2-5-3/index.xml': 7.388,
    'G2-4-2_2/index.xml': 7.564,
    'G2-5-4_1/index.xml': 7.766,
    'G2-5-4_3/index.xml': 7.796,
    'G2-4-1_2/index.xml': 7.832,
    'G3-2-5/index.xml': 7.84,
    'G3-5-1/index.xml': 7.865,
    'RTS_Annex_IV_Par_4_2/index.xml': 7.881,
    'G2-4-2_1/index.xml': 7.961,
    'G3-2-1/index.xml': 7.973,
    'G3-4-2_1/index.xml': 8.012,
    'RTS_Annex_II_Par_2/index.xml': 8.048,
    'RTS_Annex_IV_Par_1_G2-1-4/index.xml': 8.21,
    'G2-7-1_1/index.xml': 8.273,
    'G2-7-1_2/index.xml': 8.288,
    'G3-4-4/index.xml': 8.315,
    'G3-2-2/index.xml': 8.366,
    'G2-6-2/index.xml': 8.368,
    'G3-4-2_2/index.xml': 8.397,
    'G3-4-5_2/index.xml': 8.44,
    'G3-4-5_1/index.xml': 8.491,
    'G2-4-1_3/index.xml': 8.516,
    'G3-2-3/index.xml': 8.751,
    'G3-4-7/index.xml': 8.782,
    'G3-4-2_3/index.xml': 8.977,
    'RTS_Annex_IV_Par_8/index.xml': 9.929,
    'G2-5-1_3/index.xml': 10.567,
    'G3-4-3_2/index.xml': 11.164,
    'RTS_Annex_IV_Par_2_G2-1-1/index.xml': 11.229,
    'G3-4-3_1/index.xml': 11.231,
    'G2-6-1_2/index.xml': 11.548,
    'RTS_Annex_IV_Par_11_G3-2-2/index.xml': 11.728,
    'G3-4-2_4/index.xml': 11.803,
    'RTS_Annex_III_Par_1/index.xml': 12.161,
    'RTS_Annex_II_Par_1/index.xml': 12.197,
    'G2-6-1_1/index.xml': 12.314,
    'RTS_Art_3/index.xml': 12.818,
    'G3-1-1_2/index.xml': 13.467,
    'G2-2-1/index.xml': 13.522,
    'G2-1-3_1/index.xml': 13.528,
    'G2-1-3_2/index.xml': 13.635,
    'G2-2-2/index.xml': 13.687,
    'G2-3-1_2/index.xml': 13.766,
    'RTS_Annex_IV_Par_12_G2-2-4/index.xml': 14.174,
    'G2-5-1_2/index.xml': 15.168,
    'G2-5-2/index.xml': 15.286,
    'G2-5-1_1/index.xml': 15.782,
    'RTS_Annex_IV_Par_5_G3-4-6/index.xml': 15.947,
    'RTS_Annex_IV_Par_4_G1-1-1_G3-4-5/index.xml': 17.069,
    'G2-4-1_1/index.xml': 21.138,
    'G2-1-2/index.xml': 21.157,
    'RTS_Annex_IV_Par_9_Par_10_G1-4-1_G1-4-2_G3-3-1_G3-3-2/index.xml': 23.101,
    'G2-3-1_1/index.xml': 23.223,
    'G2-3-1_3/index.xml': 24.996,
    'G3-1-2/index.xml': 25.644,
    'G3-1-1_1/index.xml': 26.85,
    'G2-2-3/index.xml': 28.098,
    'G3-1-5/index.xml': 30.875,
}.items()}

config = ConformanceSuiteConfig(
    approximate_relative_timing=TIMING,
    args=[
        '--disclosureSystem', 'esef',
        '--formula', 'run',
    ],
    file='esef_conformance_suite_2021/esef_conformance_suite_2021/index_inline_xbrl.xml',
    info_url='https://www.esma.europa.eu/document/conformance-suite-2021',
    local_filepath='esef_conformance_suite_2021.zip',
    name=PurePath(__file__).stem,
    plugins=frozenset({'validate/ESEF'}),
    public_download_url='https://www.esma.europa.eu/sites/default/files/library/esef_conformance_suite_2021.zip',
    shards=5,
)



'''
This plug-in removes xmlns="http://www.w3.org/1999/xhtml" from
escaped html in text content of expected instance facts for inline XBRL text facts

(c) Copyright 2019 Mark V Systems Limited, All rights reserved.
'''
try:
    import regex as re
except ImportError:
    import re
from arelle.XhtmlValidate import htmlEltUriAttrs, resolveHtmlUri
from arelle.ValidateXbrlCalcs import ValidateCalcsMode as CalcsMode

def testcaseVariationLoaded(testInstance, testcaseInstance, modelTestcaseVariation):
    for result in modelTestcaseVariation.iter("{*}result"):
        for n, v in result.attrib.items():
            if n.endswith("mode"):
                if v == "round-to-nearest":
                    testInstance.modelManager.validateCalcs = CalcsMode.ROUND_TO_NEAREST
                elif v == "truncate":
                    testInstance.modelManager.validateCalcs = CalcsMode.TRUNCATION

__pluginInfo__ = {
    'name': 'Testcase obtain expected calc 11 mode from variation/result@mode',
    'version': '0.9',
    'description': "This plug-in removes xxx.  ",
    'license': 'Apache-2',
    'author': 'Mark V Systems Limited',
    'copyright': '(c) Copyright 2019 Mark V Systems Limited, All rights reserved.',
    # classes of mount points (required)
    'TestcaseVariation.Xbrl.Loaded': testcaseVariationLoaded,
}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from abc import ABC
from enum import Enum
from typing import Any, TYPE_CHECKING


if TYPE_CHECKING:
    import io
    from optparse import OptionParser
    from tkinter import Menu

    from arelle.Cntlr import Cntlr
    from arelle.CntlrCmdLine import CntlrCmdLine
    from arelle.CntlrWinMain import CntlrWinMain
    from arelle.DisclosureSystem import DisclosureSystem
    from arelle.FileSource import FileSource
    from arelle.ModelDocument import LoadingException, ModelDocument
    from arelle.ModelManager import ModelManager
    from arelle.ModelXbrl import ModelXbrl
    from arelle.ValidateXbrl import ValidateXbrl
    from arelle.formula.XPathContext import XPathContext
    from arelle.webserver.bottle import Bottle


class ValidationHook(Enum):
    """
    These hooks are called at different stages of validation, but all provide a common interface (ValidateXbrl is the first param).
    """

    XBRL_START = "Validate.XBRL.Start"
    XBRL_FINALLY = "Validate.XBRL.Finally"
    XBRL_DTS_DOCUMENT = "Validate.XBRL.DTS.document"
    FINALLY = "Validate.Finally"


class PluginHooks(ABC):
    @staticmethod
    def cntlrCmdLineOptions(
        parser: OptionParser,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `CntlrCmdLine.Options`

        This hook is used to add plugin specific options to the command line.

        Example:
        ```python
        parser.add_option(
            "--my-option",
            dest="myOption",
            help="My custom option",
        )
        ```

        :param parser: the parser class to add options to.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def cntlrInit(
        cntlr: Cntlr,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `Cntlr.Init`

        This hook is called while the [Cntlr](#arelle.Cntlr.Cntlr) is being initialized and after logging has been setup.

        :param cntlr: The [Cntlr](#arelle.Cntlr.Cntlr) being initialized.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def cntlrWebMainStartWebServer(
        app: Bottle,
        cntlr: CntlrCmdLine,
        host: str,
        port: str,
        server: str,
        *args: Any,
        **kwargs: Any,
    ) -> str | None:
        """
        Plugin hook: `CntlrWebMain.StartWebServer`

        This hook is for adding routes to the webserver.

        Only routes from a single plugin will be applied.

        Example:
        ```python
        app.route('/rest/my-test', "GET", my_test)
        app.route('/rest/my-run/<file:path>', ("GET", "POST"), my_run)
        ```

        :param app: The [Bottle](#arelle.webserver.bottle.Bottle) server.
        :param cntlr: The [controller](#arelle.CntlrCmdLine.CntlrCmdLine) for the server.
        :param host: The webserver host.
        :param port: The webserver port.
        :param server: The webserver path.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: `None`, a string of `skip-routes` or `skip-run`.
            Example:
            ```python
            # Block default Arelle routes.
            return "skip-routes"

            # Block default webserver startup.
            return "skip-run"

            # Block default Arelle routes and webserver startup.
            return "skip-routes|skip-run"

            # Normal routes will be combined with plugin routes and app started.
            return None
            ```
        """
        raise NotImplementedError

    @staticmethod
    def cntlrWinMainMenuHelp(
        cntlr: CntlrWinMain,
        menu: Menu,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `CntlrWinMain.Menu.Help`

        Hook for adding items to the Arelle GUI help menu.

        Example:
        ```python
        menu.add_command(label="Plugin documentation URL", command=self.openHelpDocumentation)
        ```

        :param cntlr: The [CntlrWinMain](#arelle.CntlrWinMain.CntlrWinMain) instance that the request is coming from.
        :param menu: The tkinter help menu to extend.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def cntlrWinMainMenuTools(
        cntlr: CntlrWinMain,
        menu: Menu,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `CntlrWinMain.Menu.Tools`

        Hook for adding items to the Arelle GUI tools menu.

        Example:
        ```python
        menu.add_command(label="Plugin Option", command=self.doSomething)
        ```

        :param cntlr: The [CntlrWinMain](#arelle.CntlrWinMain.CntlrWinMain) instance that the request is coming from.
        :param menu: The tkinter tools menu to extend.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def cntlrWinMainMenuValidation(
        cntlr: CntlrWinMain,
        menu: Menu,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `CntlrWinMain.Menu.Validation`

        Hook for adding items to the Arelle GUI validation menu.

        Example:
        ```python
        menu.add_command(label="Plugin validation option", command=self.validationOptions)
        ```

        :param cntlr: The [CntlrWinMain](#arelle.CntlrWinMain.CntlrWinMain) instance that the request is coming from.
        :param menu: The tkinter validation menu to extend.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def cntlrWinMainMenuView(
        cntlr: CntlrWinMain,
        menu: Menu,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `CntlrWinMain.Menu.View`

        Hook for adding items to the Arelle GUI view menu.

        Example:
        ```python
        menu.add_command(label="My Plugin Option", command=self.doSomething)
        ```

        :param cntlr: The [CntlrWinMain](#arelle.CntlrWinMain.CntlrWinMain) instance that the request is coming from.
        :param menu: The tkinter view menu to extend.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def disclosureSystemConfigURL(
        disclosureSystem: DisclosureSystem,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        """
        Plugin hook: `DisclosureSystem.ConfigURL`

        For validation plugins this provides a path to the disclosure system config file.
        See arelle/config/disclosuresystems.xml for examples.

        :param disclosureSystem: The [DisclosureSystem](#arelle.DisclosureSystem.DisclosureSystem)
               instance that the config will be applied to.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: A relative path to a disclosure system XML config file.
            Example:
            ```python
            return str(Path(__file__).parent / "resources" / "config.xml")
            ```
        """
        raise NotImplementedError

    @staticmethod
    def disclosureSystemTypes(
        disclosureSystem: DisclosureSystem,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[tuple[str, str], ...]:
        """
        Plugin hook: `DisclosureSystem.Types`

        For validation plugins this should return a tuple of tuples containing:
        1. a validation type (matching a validation type from a disclosure system config.
        2. a test type property that's applied to the [DisclosureSystem](#arelle.DisclosureSystem.DisclosureSystem).
           It will be set to `True` if the disclosure system for the validation type above is
           selected and `false` otherwise.

        :param disclosureSystem: The [DisclosureSystem](#arelle.DisclosureSystem.DisclosureSystem)
               instance that the types will be applied to.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: A tuple of two item tuples containing a validation type and test property.
            Example:
            ```python
            return (("ESEF", "ESEFplugin"),)
            ```
        """
        raise NotImplementedError

    @staticmethod
    def fileSourceExists(
        cntlr: Cntlr,
        filepath: str,
        *args: Any,
        **kwargs: Any,
    ) -> bool | None:
        """
        Plugin hook: `FileSource.Exists`

        This hook can be used to override [FileSource](#arelle.FileSource.FileSource) existence checks.
        For instance for a plugin that encrypts files, creating a copy with a different file extension,
        and the filepath should be transformed accordingly.

        :param cntlr: The [controller](#arelle.Cntlr.Cntlr) the current request is running from.
        :param filepath: The path which Arelle is checking for the existence of a file.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None or boolean value indicating if the file exists.
            Example:
            ```python
            # Defer to other plugins and/or default behavior.
            return None
            # Indicate file exists.
            return True
            # Indicate file does not exist.
            return False
            ```
        """
        raise NotImplementedError

    @staticmethod
    def fileSourceFile(
        cntlr: Cntlr,
        filepath: str,
        binary: bool,
        stripDeclaration: bool,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[io.BytesIO | None] | tuple[io.TextIOWrapper, str]:
        """
        Plugin hook: `FileSource.File`

        fileResult = pluginMethod(self.cntlr, filepath, binary, stripDeclaration)

        This hook can be used to override [FileSource](#arelle.FileSource.FileSource) file open operations.

        :param cntlr: The [controller](#arelle.Cntlr.Cntlr) the current request is running from.
        :param filepath: The path of the file to open.
        :param binary: Indicates if the file should be opened as binary.
        :param stripDeclaration: Indicates if XML declaration should be stripped.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: tuple of text IO and encoding, or binary io and None
            Example:
            ```python
            # Binary IO
            return (io.BytesIO(b),)
            # Text IO
            return return (io.TextIOWrapper(io.BytesIO(b), encoding="utf-8"), "utf-8")
            ```
        """
        raise NotImplementedError

    @staticmethod
    def modelDocumentIsPullLoadable(
        modelXbrl: ModelXbrl,
        mappedUri: str,
        normalizedUri: str,
        filepath: str,
        isEntry: bool,
        namespace: str,
        *args: Any,
        **kwargs: Any,
    ) -> bool:
        """
        Plugin hook: `ModelDocument.IsPullLoadable`

        Plugin hook to signal files which can be opened by the plugin that otherwise would not be supported by Arelle.
        For instance a plugin that enables opening and loading data from json files.

        :param modelXbrl: The [ModelXbrl](#arelle.ModelXbrl.ModelXbrl) being constructed.
        :param mappedUri: The path of the document. May be the same as the `normalizedUri` or a path to a file in a taxonomy package.
        :param normalizedUri: The normalized path of the document. This can be a URL or local file system path.
        :param filepath: The path of the document. May be the same as the `normalizedUri` or a path to a file in the local cache.
        :param isEntry: Boolean value indicating if the document is an entrypoint.
        :param namespace: The target namespace if the document has one.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: `True` if the plugin can load the file or `False` if not.
            Example:
            ```python
            return Path(filepath).suffix == ".json"
            ```
        """
        raise NotImplementedError

    @staticmethod
    def modelDocumentPullLoader(
        modelXbrl: ModelXbrl,
        normalizedUri: str,
        filepath: str,
        isEntry: bool,
        namespace: str | None,
        *args: Any,
        **kwargs: Any,
    ) -> ModelDocument | LoadingException | None:
        """
        Plugin hook: `ModelDocument.PullLoader`

        This plugin hook can be used to override [ModelDocument](#arelle.ModelDocument.ModelDocument) construction.

        For instance for a plugin that enables opening and loading data from Excel files.

        This hook can also be used to log an initial error and/or return a
        [LoadingException](#arelle.ModelDocument.LoadingException)
        if the XBRL report doesn't match a naming requirement.

        :param modelXbrl: The [ModelXbrl](#arelle.ModelXbrl.ModelXbrl) being constructed.
        :param normalizedUri: The normalized path of the document. This can be a URL or local file system path.
        :param filepath: The path of the document. May be the same as the `normalizedUri` or a path to a file in the local cache.
        :param isEntry: Boolean value indicating if the document is an entrypoint.
        :param namespace: The target namespace if the document has one.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: A [ModelDocument](#arelle.ModelDocument.ModelDocument), a
                 [LoadingException](#arelle.ModelDocument.LoadingException], or `None`.
            Example:
            ```python
            # Defer to other plugins and/or default behavior.
            return None

            # Override ModelDocument construction from plugin.
            return ModelDocument(...)

            # Signal that document can't be constructed.
            return LoadingException("Invalid document")
            ```
        """
        raise NotImplementedError

    @staticmethod
    def modelManagerLoad(
        manager: ModelManager,
        fileSource: FileSource,
        *args: Any,
        **kwargs: Any,
    ) -> ModelXbrl | None:
        """
        Plugin hook: `ModelManager.Load`

        Hook to override default [ModelXbrl](#arelle.ModelXbrl.ModelXbrl) loading.
        Return `None` to defer to other plugins of default loading.

        :param manager: The [ModelManager](#arelle.ModelManager.ModelManager) the request is coming from.
        :param fileSource: The [FileSource](#arelle.FileSource.FileSource) to load.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: Either a [ModelXbrl](#arelle.ModelXbrl.ModelXbrl) instance or `None` to defer to default loading.
        """
        raise NotImplementedError

    @staticmethod
    def modelXbrlInit(
        modelXbrl: ModelXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `ModelXbrl.Init`

        This plugin hook is called as the last step of the [ModelXbrl](#arelle.ModelXbrl.ModelXbrl)
        constructor which is prior to [ModelDocument](#arelle.ModelDocument.ModelDocument) loading.

        :param modelXbrl: The [XBRL model](#arelle.ModelXbrl.ModelXbrl) being constructed.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def modelXbrlLoadComplete(
        modelXbrl: ModelXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `ModelXbrl.LoadComplete`

        This plugin hook is called after the [ModelXbrl](#arelle.ModelXbrl.ModelXbrl) and
        [ModelDocument](#arelle.ModelDocument.ModelDocument) loading is complete.

        :param modelXbrl: The constructed [ModelXbrl](#arelle.ModelXbrl.ModelXbrl).
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def validateFinally(
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `Validate.Finally`

        This plugin hook is called after formula processing and other XBRL validation.
        This hook is useful for logging errors or warnings based on other errors or warnings that were triggered.

        Example:
        ```python
        if "XYZ.01.01" in val.modelXbrl.errors and "XYZ.21.02" not in val.modelXbrl.errors:
            modelXbrl.error("XYZ.52.04", "Incompatible data reported.")
        ```

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def validateFormulaCompiled(
        modelXbrl: ModelXbrl,
        xpathContext: XPathContext,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `ValidateFormula.Compiled`

        Hook executed after formula rules are compiled, but before they're executed.

        :param modelXbrl: The [ModelXbrl](#arelle.ModelXbrl.ModelXbrl).
        :param xpathContext: The formula evaluation [XPathContext](#arelle.formula.XPathContext.XPathContext).
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def validateFormulaFinished(
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `ValidateFormula.Finished`

        Hook executed after formula evaluation.

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def validateXbrlDtsDocument(
        val: ValidateXbrl,
        modelDocument: ModelDocument,
        isFilingDocument: bool,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `Validate.XBRL.DTS.document`

        Validation hook for implementing DTS validation rules.
        Useful for implementing validation rules related to extension taxonomies.

        Example:
        ```python
        if (
            modelDocument.type == ModelDocument.Type.SCHEMA
            and modelDocument.targetNamespace is not None
            and len(modelDocument.targetNamespace) > 100
        ):
            val.modelXbrl.error(
                codes="XYZ.1.9.13",
                msg="TargetNamespace is too long.",
            )
        ```

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param modelDocument: The [ModelDocument](#arelle.ModelDocument.ModelDocument) to validate.
        :param isFilingDocument: Indicates if document is part of filing or external reference.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def validateXbrlFinally(
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `Validate.XBRL.Finally`

        Hook for executing validation rules after other XBRL validations, but prior to formula processing.

        Example:
        ```python
        if "Cash" not in val.modelXbrl.factsByLocalName:
            val.modelXbrl.error(codes="01.01", msg="Cash must be reported.")
        if "Assets" not in val.modelXbrl.factsByLocalName:
            val.modelXbrl.warning(codes=("01.02", "13.04"), msg="Assets should be reported.")
        ```

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError

    @staticmethod
    def validateXbrlStart(
        val: ValidateXbrl,
        parameters: dict[Any, Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Plugin hook: `Validate.XBRL.Start`

        Hook for executing validation rules prior to other XBRL validations.

        Example:
        ```python
        if "Cash" not in val.modelXbrl.factsByLocalName:
            val.modelXbrl.error(codes="01.01", msg="Cash must be reported.")
        if "Assets" not in val.modelXbrl.factsByLocalName:
            val.modelXbrl.warning(codes=("01.02", "13.04"), msg="Assets should be reported.")
        ```

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param parameters: dictionary of validation parameters.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        raise NotImplementedError



'''
This module implements Quick Books server mode

See COPYRIGHT.md for copyright information.
'''
from lxml import etree
import uuid, io, datetime
from arelle import XmlUtil

clientVersion = None
userName = None
sessions = {}  # use when interactive session started by Quickbooks side (not used now)
qbRequests = []  # used by rest API or GUI requests for QB data
qbRequestStatus = {}
xbrlInstances = {}
cntlr = None

# report in url path request and type of query to QB
supportedQbReports = {'trialBalance':'GeneralSummary',
                      'generalLedger':'GeneralDetail',
                      'journal':'GeneralDetail'
                     }
# some reports don't provide the needed columns, request explicitly
includeQbColumns = {'trialBalance': '',
                    'generalLedger': '''
<IncludeColumn>TxnType</IncludeColumn>
<IncludeColumn>Date</IncludeColumn>
<IncludeColumn>RefNumber</IncludeColumn>
<IncludeColumn>Name</IncludeColumn>
<IncludeColumn>Memo</IncludeColumn>
<IncludeColumn>SplitAccount</IncludeColumn>
<IncludeColumn>Credit</IncludeColumn>
<IncludeColumn>Debit</IncludeColumn>
<IncludeColumn>RunningBalance</IncludeColumn>
''',
                 'journal': ''
                 }
glEntriesType = {'trialBalance':'trialbalance',
                 'generalLedger':'balance',
                 'journal':'journal'
                }

qbTxnTypeToGL = {# QB code is case insensitive comparision (lowercase, some QBs do not have expected camel case)
                 'bill':'voucher', # bills from vendors
                 'billpayment':'check', # credits from vendors
                 'billpaymentcheck':'check', # payments to vendors from bank account
                 'billpmt-check':'check',  # QB 2009
                 'billpaymentcreditcard':'payment-other',  # payments to vendor from credit card account
                 'buildassembly':'other',
                 'charge':'other',
                 'check':'check', # checks written on bank account
                 'credit':'credit-memo',
                 'creditcardcharge':'payment-other', # credit card account charge
                 'creditcardcredit':'other', # credit card account credit
                 'creditmemo':'credit-memo', # credit memo to customer
                 'deposit':'check', # GL calls it check whether sent or received
                 'discount':'credit-memo',
                 'estimate':'other',
                 'generaljournal':'manual-adjustment',
                 'inventoryadjustment':'other',
                 'invoice':'invoice',
                 'itemreceipt':'receipt',
                 'journalentry':'manual-adjustment',
                 'liabilitycheck': 'check',
                 'payment': 'check',
                 'paycheck': 'check',
                 'purchaseorder':'order-vendor',
                 'receivepayment':'payment-other',
                 'salesorder':'order-customer',
                 'salesreceipt':'other',
                 'salestaxpaymentcheck':'check',
                 'statementcharge':'other',
                 'transfer':'payment-other',
                 'vendorcredit':'credit-memo',
                 }

def server(_cntlr, soapFile, requestUrlParts):
    global cntlr
    if cntlr is None: cntlr = _cntlr
    soapDocument = etree.parse(soapFile)
    soapBody = soapDocument.find("{http://schemas.xmlsoap.org/soap/envelope/}Body")
    if soapBody is None:
        return ""
    else:
        for request in soapBody.iterchildren():
            requestName = request.tag.partition("}")[2]
            print ("request {0}".format(requestName))
            response = None
            if request.tag == "{http://developer.intuit.com/}serverVersion":
                response = "Arelle 1.0"
            elif request.tag == "{http://developer.intuit.com/}clientVersion":
                global clientVersion
                clientVersion = request.find("{http://developer.intuit.com/}strVersion").text
            elif request.tag == "{http://developer.intuit.com/}authenticate":
                #global userName  # not needed for now
                #userName = request.find("{http://developer.intuit.com/}strUserName").text
                #password is ignored
                ticket = str(uuid.uuid1())
                global qbRequests
                if qbRequests: # start a non-interactive session
                    response = [ticket, ""]
                    sessions[ticket] = qbRequests
                    qbRequests = []
                else:
                    # to start an interactive session automatically from QB side, uncomment
                    #response = [ticket, "" if not sessions else "none"] # don't start session if one already there
                    #sessions[ticket] = [{"request":"StartInteractiveMode"}]
                    response = [ticket, "none"]  # response to not start interactive mode
            elif request.tag == "{http://developer.intuit.com/}sendRequestXML":
                ticket = request.find("{http://developer.intuit.com/}ticket").text
                _qbRequests = sessions.get(ticket)
                if _qbRequests:
                    _qbRequest = _qbRequests[0]
                    action = _qbRequest["request"]
                    if action == "StartInteractiveMode":
                        response = ''
                    elif action in supportedQbReports:
                        # add company info to request dict
                        _qbRequest["strHCPResponse"] = request.find("{http://developer.intuit.com/}strHCPResponse").text
                        response = ('''<?xml version="1.0"?>
<?qbxml version="8.0"?>
<QBXML>
  <QBXMLMsgsRq onError="stopOnError">
    <{1}ReportQueryRq>
      <{1}ReportType>{0}</{1}ReportType>
      <ReportPeriod>
        <FromReportDate>{2}</FromReportDate>
        <ToReportDate>{3}</ToReportDate>
      </ReportPeriod>{4}
    </{1}ReportQueryRq>
  </QBXMLMsgsRq>
</QBXML>''').format(action[0].upper() + action[1:],
                    supportedQbReports[action],
                    _qbRequest["fromDate"],
                    _qbRequest["toDate"],
                    includeQbColumns[action],
                    ).replace("&","&amp;").replace("<","&lt;").replace(">","&gt;")

            elif request.tag == "{http://developer.intuit.com/}connectionError":
                ticket = request.find("{http://developer.intuit.com/}ticket").text
                hresult = request.find("{http://developer.intuit.com/}hresult").text
                if hresult and hresult.startswith("0x"):
                    hresult = hresult[2:] # remove 0x if present
                message = request.find("{http://developer.intuit.com/}message").text
                print ("connection error message: [{0}] {1}".format(hresult, message))
                _qbRequests = sessions.get(ticket)
                if _qbRequests:
                    qbRequestTicket = _qbRequests[0]["ticket"]
                    qbRequestStatus[qbRequestTicket] = "ConnectionErrorMessage: [{0}] {1}".format(hresult, message)
                response = "done"
            elif request.tag == "{http://developer.intuit.com/}receiveResponseXML":
                ticket = request.find("{http://developer.intuit.com/}ticket").text
                responseXml = (request.find("{http://developer.intuit.com/}response").text or "").replace("&lt;","<").replace("&gt;",">")
                _qbRequests = sessions.get(ticket)
                if _qbRequests:
                    if responseXml:
                        processQbResponse(_qbRequests[0], responseXml)
                    else:
                        print ("no response from QuickBooks")
                    response = str(100 / len(_qbRequests))
                    sessions[ticket] = _qbRequests[1:]
            elif request.tag == "{http://developer.intuit.com/}getLastError":
                ticket = request.find("{http://developer.intuit.com/}ticket").text
                _qbRequests = sessions.get(ticket)
                if _qbRequests:
                    _qbRequest = _qbRequests[0]
                    action = _qbRequest["request"]
                    if action == "StartInteractiveMode":
                        response = "Interactive mode"
                    else:
                        response = "NoOp"
                else:
                    response = "NoOp"
            elif request.tag == "{http://developer.intuit.com/}getInteractiveURL":
                ticket = request.find("{http://developer.intuit.com/}wcTicket").text
                response = "{0}://{1}/quickbooks/server.html?ticket={2}".format(
                            requestUrlParts.scheme,
                            requestUrlParts.netloc,
                            ticket)
                sessions[ticket] = [{"request":"WaitForInput"}]
            elif request.tag == "{http://developer.intuit.com/}isInteractiveDone":
                ticket = request.find("{http://developer.intuit.com/}wcTicket").text
                _qbRequests = sessions.get(ticket)
                if _qbRequests:
                    _qbRequest = _qbRequests[0]
                    action = _qbRequest["request"]
                    if action == "Done":
                        response = "Done"
                    else:
                        response = "Not done"
                else:
                    response = "Not done"
            elif request.tag == "{http://developer.intuit.com/}interactiveRejected":
                ticket = request.find("{http://developer.intuit.com/}wcTicket").text
                response = "Interactive session timed out or canceled"
                sessions.pop(ticket, None)
            elif request.tag == "{http://developer.intuit.com/}closeConnection":
                response = "OK"

            soapResponse = qbResponse(requestName, response)
            return soapResponse

def qbRequest(qbReport, fromDate, toDate, file):
    ticket = str(uuid.uuid1())
    qbRequests.append({"ticket":ticket,
                       "request":qbReport,
                       "fromDate":fromDate,
                       "toDate":toDate,
                       "xbrlFile":file})
    qbRequestStatus[ticket] = _("Waiting for QuickBooks")
    return ticket

def qbResponse(responseName, content=None):
    if not content:
        result = ""
    elif isinstance(content, list):
        result = '<{0}Result>{1}</{0}Result>'.format(
                 responseName,
                 '\n'.join("<string>{0}</string>".format(l) for l in content))
    else:
        result = '<{0}Result>{1}</{0}Result>'.format(responseName, content)

    return ('<?xml version="1.0" encoding="utf-8"?>'
            '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema">'
            '<soap:Body>'
            '<{0}Response xmlns="http://developer.intuit.com/">'
            '{1}'
            '</{0}Response>'
            '</soap:Body>'
            '</soap:Envelope>'.format(responseName, result))

def docEltText(doc, tag, defaultValue=""):
    for elt in doc.iter(tag):
        return elt.text
    return defaultValue

def processQbResponse(qbRequest, responseXml):
    from arelle import ModelXbrl, XbrlConst
    from arelle.ModelValue import qname
    ticket = qbRequest["ticket"]
    qbRequestStatus[ticket] = _("Generating XBRL-GL from QuickBooks response")
    qbReport = qbRequest["request"]
    xbrlFile = qbRequest["xbrlFile"]
    fromDate = qbRequest["fromDate"]
    toDate = qbRequest["toDate"]
    strHCPResponse = qbRequest.get("strHCPResponse", "")

    # uncomment to dump out QB responses
    '''
    with open("c:/temp/test.xml", "w") as fh:
        fh.write(responseXml)
    with open("c:/temp/testC.xml", "w") as fh:
        fh.write(strHCPResponse)
    # qb responses dump
    '''

    companyQbDoc = etree.parse(io.StringIO(initial_value=strHCPResponse))
    responseQbDoc = etree.parse(io.StringIO(initial_value=responseXml))
    # columns table
    colTypeId = {}
    colIdType = {}
    for colDescElt in responseQbDoc.iter("ColDesc"):
        colTypeElt = colDescElt.find("ColType")
        if colTypeElt is not None:
            colID = colDescElt.get("colID")
            colType = colTypeElt.text
            if colType == "Amount": # check if there's a credit or debit colTitle
                for colTitleElt in colDescElt.iter("ColTitle"):
                    title = colTitleElt.get("value")
                    if title in ("Credit", "Debit"):
                        colType = title
                        break
            colTypeId[colType] = colID
            colIdType[colID] = colType

    # open new result instance document

    # load GL palette file (no instance)
    instance = cntlr.modelManager.load("http://www.xbrl.org/taxonomy/int/gl/2006-10-25/plt/case-c-b-m-u-t/gl-plt-2006-10-25.xsd")
    if xbrlFile is None:
        xbrlFile = "sampleInstance.xbrl"
        saveInstance = False
    else:
        saveInstance = True
    instance.createInstance(xbrlFile) # creates an instance as this modelXbrl's entrypoing
    newCntx = instance.createContext("http://www.xbrl.org/xbrlgl/sample", "SAMPLE",
                  "instant", None, datetime.date.today() + datetime.timedelta(1), # today midnight
                  None, {}, [], [], afterSibling=ModelXbrl.AUTO_LOCATE_ELEMENT)

    monetaryUnit = qname(XbrlConst.iso4217, "iso4217:USD")
    newUnit = instance.createUnit([monetaryUnit],[], afterSibling=ModelXbrl.AUTO_LOCATE_ELEMENT)

    nonNumAttr = [("contextRef", newCntx.id)]
    monetaryAttr = [("contextRef", newCntx.id), ("unitRef", newUnit.id), ("decimals", "2")]

    isoLanguage = qname("{http://www.xbrl.org/2005/iso639}iso639:en")

    # root of GL is accounting entries tuple
    xbrlElt = instance.modelDocument.xmlRootElement

    '''The container for XBRL GL, accountingEntries, is not the root of an XBRL GL file - the root,
    as with all XBRL files, is xbrl. This means that a single XBRL GL file can store one or more
    virtual XBRL GL files, through one or more accountingEntries structures with data inside.
    The primary key to understanding an XBRL GL file is the entriesType. A single physical XBRL GL
    file can have multiple accountingEntries structures to represent both transactions and
    master files; the differences are signified by the appropriate entriesType enumerated values.'''
    accountingEntries = instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:accountingEntries"))

    # Because entriesType is strongly suggested, documentInfo will be required
    docInfo = instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:documentInfo"), parent=accountingEntries)
    # This field, entriesType, provides the automated guidance on the purpose of the XBRL GL information.
    instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:entriesType"), parent=docInfo, attributes=nonNumAttr,
                        text=glEntriesType[qbReport])
    '''Like a serial number, this field, uniqueID, provides a place to uniquely identify/track
    a series of entries. It is like less relevant for ad-hoc reports. XBRL GL provides for later
    correction through replacement or augmentation of transferred information.'''
    instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:uniqueID"), parent=docInfo, attributes=nonNumAttr,
                        text="001")
    instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:language"), parent=docInfo, attributes=nonNumAttr,
                        text=XmlUtil.addQnameValue(xbrlElt, isoLanguage))
    '''The date associated with the creation of the data reflected within the associated
    accountingEntries section. Somewhat like a "printed date" on a paper report'''
    instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:creationDate"), parent=docInfo, attributes=nonNumAttr,
                        text=str(datetime.date.today()))
    instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:periodCoveredStart"), parent=docInfo, attributes=nonNumAttr,
                        text=fromDate)
    instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:periodCoveredEnd"), parent=docInfo, attributes=nonNumAttr,
                        text=toDate)
    instance.createFact(qname("{http://www.xbrl.org/int/gl/bus/2006-10-25}gl-bus:sourceApplication"), parent=docInfo, attributes=nonNumAttr,
                        text=docEltText(companyQbDoc, "ProductName","QuickBooks (version not known)"))
    instance.createFact(qname("{http://www.xbrl.org/int/gl/muc/2006-10-25}gl-muc:defaultCurrency"), parent=docInfo, attributes=nonNumAttr,
                        text=XmlUtil.addQnameValue(xbrlElt, monetaryUnit))

    '''Typically, an export from an accounting system does not carry with it information
    specifically about the company. However, the name of the company would be a very good
    thing to include with the file, making the entityInformation tuple necessary.'''
    entityInfo = instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:entityInformation"), parent=accountingEntries)
    '''The name of the company would be a very good thing to include with the file;
    this structure and its content are where that would be stored.'''
    orgIds = instance.createFact(qname("{http://www.xbrl.org/int/gl/bus/2006-10-25}gl-bus:organizationIdentifiers"), parent=entityInfo)
    instance.createFact(qname("{http://www.xbrl.org/int/gl/bus/2006-10-25}gl-bus:organizationIdentifier"), parent=orgIds, attributes=nonNumAttr,
                        text=docEltText(companyQbDoc, "CompanyName"))
    instance.createFact(qname("{http://www.xbrl.org/int/gl/bus/2006-10-25}gl-bus:organizationDescription"), parent=orgIds, attributes=nonNumAttr,
                        text=docEltText(companyQbDoc, "LegalCompanyName"))

    if qbReport == "trialBalance":
        qbTxnType = "trialbalance"
    else:
        qbTxnType = None
    qbTxnNumber = None
    qbDate = None
    qbRefNumber = None
    isFirst = True
    entryNumber = 1
    lineNumber = 1

    for dataRowElt in responseQbDoc.iter("DataRow"):
        cols = dict((colIdType[colElt.get("colID")], colElt.get("value")) for colElt in dataRowElt.iter("ColData"))
        if qbReport == "trialBalance" and "Label" in cols:
            cols["SplitAccount"] = cols["Label"]

        hasRowDataAccount = False
        for rowDataElt in dataRowElt.iter("RowData"):
            rowType = rowDataElt.get("rowType")
            if rowType == "account":
                hasRowDataAccount = True
                if "SplitAccount" not in cols:
                    cols["SplitAccount"] = rowDataElt.get("value")
        if qbReport == "trialBalance" and not hasRowDataAccount:
            continue  # skip total lines or others without account information
        elif qbReport in ("generalLedger", "journal"):
            if "TxnType" not in cols:
                continue  # not a reportable entry

        # entry header fields only on new item that generates an entry header
        if "TxnType" in cols:
            qbTxnType = cols["TxnType"]
        if "TxnNumber" in cols:
            qbTxnNumber = cols["TxnNumber"]
        if "Date" in cols:
            qbDate = cols["Date"]
        if "RefNumber" in cols:
            qbRefNumber = cols["RefNumber"]
        # entry details provided on every entry
        qbName = cols.get("Name")
        qbMemo = cols.get("Memo")
        qbAccount = cols.get("SplitAccount")
        qbAmount = cols.get("Amount")
        qbDebitAmount = cols.get("Debit")
        qbCreditAmount = cols.get("Credit")
        runningBalance = cols.get("RunningBalance")

        if qbAmount is not None:
            drCrCode = None
            amt = qbAmount
        elif qbDebitAmount is not None:
            drCrCode = "D"
            amt = qbDebitAmount
        elif qbCreditAmount is not None:
            drCrCode = "C"
            amt = qbCreditAmount
        else:
            # no amount, skip this transaction
            continue

        if isFirst or qbTxnNumber:
            '''Journal entries require entry in entryHeader and entryDetail.
            Few files can be represented using only documentInfo and entityInformation sections,
            but it is certainly possible.'''
            entryHdr = instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:entryHeader"), parent=accountingEntries)
            #instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:enteredBy"), parent=entryHdr, attributes=nonNumAttr, text="")
            instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:enteredDate"), parent=entryHdr, attributes=nonNumAttr,
                                text=str(datetime.date.today()))
            '''This is an enumerated entry that ties the source journal from the reporting
            organization to a fixed list that helps in data interchange.'''
            instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:sourceJournalID"), parent=entryHdr, attributes=nonNumAttr,
                                text="gj")
            '''Since sourceJournalID is enumerated (you must pick one of the entries already
            identified within XBRL GL), sourceJournalDescription lets you capture the actual
            code or term used to descibe the source journal by the organization.'''
            # instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:sourceJournalDescription"), parent=entryHdr, attributes=nonNumAttr, text="JE")
            '''An enumerated field to differentiate between details that represent actual accounting
            entries - as opposed to entries for budget purposes, planning purposes, or other entries
            that may not contribute to the financial statements.'''
            instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:entryType"), parent=entryHdr, attributes=nonNumAttr,
                                text="standard")
            '''When capturing journal entries, you have a series of debits and credits that (normally)
            add up to zero. The hierarchical nature of XBRL GL keeps the entry detail lines associated
            with the entry header by a parent-child relationship. The unique identifier of each entry
            is entered here.'''
            instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:entryNumber"), parent=entryHdr, attributes=nonNumAttr,
                                text=str(entryNumber))
            entryNumber += 1
            # The reason for making an entry goes here.
            if qbRefNumber:
                instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:entryComment"), parent=entryHdr, attributes=nonNumAttr,
                                    text=qbRefNumber)

        '''Individual lines of journal entries will normally require their own entryDetail section -
        one primary amount per entryDetail line. However, you can list different accounts within
        the same entryDetail line that are associated with that amount. For example, if you
        capitalize for US GAAP and expense for IFRS'''
        entryDetail = instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:entryDetail"), parent=entryHdr)
        # A unique identifier for each entry detail line within an entry header, this should at the least be a counter.
        instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:lineNumber"), parent=entryDetail, attributes=nonNumAttr,
                            text=str(lineNumber))
        lineNumber += 1

        '''If account information is represented elsewhere or as a master file, some of the
        fields below would not need to be here (signified by *)'''
        account = instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:account"), parent=entryDetail)
        '''The account number is the basis for posting journal entries. In some cases,
        accounting systems used by small organizations do not use account numbers/codes,
        but only use a descriptive name for the account.'''
        # QB does not have account numbers
        # instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:accountMainID"), parent=account, attributes=nonNumAttr, text="10100")
        '''In most cases, the description is given to help a human reader; the accountMainID would
        be sufficient for data exchange purposes. As noted previously, some implementations use the
        description as the primary identifier of the account.'''
        if qbAccount:
            instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:accountMainDescription"), parent=account, attributes=nonNumAttr,
                                text=qbAccount)
        '''Accounts serve many purposes, and in a large company using more sophisticated software,
        the company may wish to record the account used for the original entry and a separate
        consolidating account. The Japanese system may require a counterbalancing account for
        each line item. And an entry may be recorded differently for US GAAP, IFRS and other purposes.
        This code is an enumerated code to help identify accounts for those purposes.'''
        instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:accountPurposeCode"), parent=account, attributes=nonNumAttr,
                            text="usgaap")
        '''In an international environment, the "chart of accounts" will include not only
        traditional accounts, like Cash, Accounts Payable/Due to Creditors or Retained Earnings,
        but also extensions to some of the accounts. Accounts Payable may be extended to
        include the creditors/vendors themselves. Therefore, in XBRL GL, accounts can be
        specifically identified as the "traditional" accountm or to identify a customer,
        vendor, employee, bank, job or fixed asset. While this may overlap with the customers,
        vendors and employees of the identifier structure, fixed-assets in the measurable
        structure, jobs in the jobInfo structure and other representations, they can also be
        represented here as appropriate to the jurisidiction.'''
        instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:accountType"), parent=account, attributes=nonNumAttr, text="account")

        '''What is a journal entry without a (monetary) amount? While XBRL GL may usher in journal
        entries that also incorporate quantities, to reflect the detail of business metrics, the
        (monetary) amount is another key and obvious fields. XBRL GL has been designed to reflect
        how popular accounting systems store amounts - some combination of a signed amount (e.g., 5, -10),
        a separate sign (entered into signOfAmount) and a separate place to indicate the number is
        associated with a debit or credit (debitCreditCode).'''
        instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:amount"), parent=entryDetail, attributes=monetaryAttr,
                            text=amt)
        '''Depending on the originating system, this field may contain whether the amount is
        associated with a debit or credit. Interpreting the number correctly for import requires
        an understanding of the three related amount fields - amount, debitCreditCode and sign of amount.'''
        if drCrCode:
            instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:debitCreditCode"), parent=entryDetail, attributes=nonNumAttr,
                                text=drCrCode)
        '''Depending on the originating system, this field may contain whether the amount is
        signed (+ or -) separately from the amount field itself. Interpreting the number correctly
        for import requires an understanding of the three related amount fields - amount,
        debitCreditCode and sign of amount.'''
        # instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:signOfAmount"), parent=entryDetail, attributes=nonNumAttr, text="+")
        # This date is the accounting significance date, not the date that entries were actually entered or posted to the system.
        if qbDate:
            instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:postingDate"), parent=entryDetail, attributes=nonNumAttr,
                                text=qbDate)

        if qbName or qbMemo:
            identRef = instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:identifierReference"), parent=entryDetail)
            if qbMemo:
                instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:identifierCode"), parent=identRef, attributes=nonNumAttr,
                                    text=qbMemo)
            if qbName:
                instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:identifierDescription"), parent=identRef, attributes=nonNumAttr,
                                    text=qbName)
            #instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:identifierType"), parent=identRef, attributes=nonNumAttr,
            #                    text="V")

        if qbReport != "trialBalance":
            if qbTxnType: # not exactly same enumerations as expected by QB
                cleanedQbTxnType = qbTxnType.replace(" ","").lower()
                glDocType = qbTxnTypeToGL.get(cleanedQbTxnType) # try table lookup
                if glDocType is None: # not in table
                    if cleanedQbTxnType.endswith("check"): # didn't convert, probably should be a check
                        glDocType = "check"
                    # TBD add more QB transations here as they are discovered and not in table
                    else:
                        glDocType = qbTxnType # if all else fails pass through QB TxnType, it will fail GL validation and be noticed!
                instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:documentType"), parent=entryDetail, attributes=nonNumAttr,
                                    text=glDocType)

            '''This enumerated field is used to specifically state whether the entries have been
            posted to the originating system or not.'''
            instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:postingStatus"), parent=entryDetail, attributes=nonNumAttr,
                                text="posted")
            # A comment at the individual entry detail level.
            # instance.createFact(qname("{http://www.xbrl.org/int/gl/cor/2006-10-25}gl-cor:detailComment"), parent=entryDetail, attributes=nonNumAttr, text="Comment...")

        isFirst = False

    if saveInstance:
        qbRequestStatus[ticket] = _("Saving XBRL-GL instance")
        instance.saveInstance()
    qbRequestStatus[ticket] = _("Done")
    # TBD resolve errors
    instance.errors = []  # TBD fix this
    xbrlInstances[ticket] = instance.uuid
